import math
from typing import Union, Optional, Sequence

import numpy
import scipy

# from numpy.typing import ArrayLike
from scipy.ndimage import gaussian_filter
from scipy.optimize import curve_fit

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lprint, lsection


class MotionStabilisationTransform(ImageTransformBase):
    """Motion Stabilisation

    Denoising is more effective if signal-correlated voxels are close to each
    other. When a 2D+t or 3D+t timelapse has shifts between time points,
    pixels that should be close to each other from one time point to the next
    are now further away from each other. Worse, this relative placement
    often varies over time. This complicates denoising and typically leads to
    more blurry denoised images. Thus, stabilizing a timelapse before
    denoising is recommended to improve denoising performance. Currently,
    we assume that all frames can be registered to a common reference frame,
    and thus that all images have a common background that can be used for
    registration. For completeness, multiple axis can be specified and the
    correction is applied along each in sequence.
    """

    preprocess_description = (
        "Stabilise motion" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Reapply motion" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = False

    def __init__(
        self,
        axes: Union[None, int, Sequence[int]] = 0,
        sigma: float = 3,
        pad: bool = False,
        crop: bool = False,
        pad_mode: str = 'min_constant',
        max_pixel_shift: Optional[int] = None,
        reference_index: Optional[int] = None,
        priority: float = 0.45,
        **kwargs,
    ):
        """
        Constructs a Motion Corrector

        Parameters
        ----------
        axes : Union[None, int, Sequence[int]]
            Index of time axis. If None the axes
            are determined automatically.
        sigma : float
            Sigma for Gaussian filtering used to
            facilitate registration.
        pad : bool
            Pads image before applying stabilisation,
            default value is True.
        crop : bool
            Crops image after undoing stabilisation,
            default value is True
        pad_mode : str
            Padding mode. Can be: 'mean_constant',
            'min_constant', 'max_constant', 'wrap'.
            Default is 'wrap'.
        max_pixel_shift : Optional[int]
            Maximum correctable motion in pixels.
            If None the maximum shift is automatically
            determined.
        reference_index : Optional[int]
            Index of image used as reference to register
            all others. If None the reference image is
            automatically determined.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.sigma = sigma
        self.axes = (
            None if axes is None else ((axes,) if isinstance(axes, int) else axes)
        )
        self.pad_mode = pad_mode

        self.pad = pad
        self.crop = crop

        self.center = False
        self.max_pixel_shift = max_pixel_shift
        self.reference_index = reference_index

        self._shifts = {}
        self._original_dtype = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_shifts']
        del state['_original_dtype']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (pad_mode={self.pad_mode},'
            f' pad={self.pad},'
            f' crop={self.crop},'
            f' center={self.center},'
            f' max_pixel_shift={self.max_pixel_shift},'
            f' reference_index={self.reference_index})'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):
        with lsection(f"Motion-correcting array of shape: {array.shape}:"):

            self._original_dtype = array.dtype

            # We need a copy because the shift-transfrom is in-place
            array = array.astype(numpy.float32, copy=True)

            axes = range(array.ndim) if self.axes is None else self.axes

            self._shifts = {}
            for axis in axes:
                lprint(f"Correcting along axis: {axis}")
                array = self._permutate(array, axis=axis)
                shifts, mean_shift = _measure_shifts(
                    array,
                    reference_index=self.reference_index,
                    center=self.center,
                    max_pixel_shift=self.max_pixel_shift,
                    mode='com',
                    sigma=self.sigma,
                )
                lprint(f"Mean shift: {mean_shift}")
                array = _shift_transform(
                    array, -shifts, pad=self.pad, crop=False, pad_mode=self.pad_mode
                )
                self._shifts[axis] = shifts
                array = self._depermutate(array, axis=axis)
            return array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(f"Undoing motion-correction for array of shape: {array.shape}:"):

            # We need a copy because the shift-transfrom is in-place
            array = array.astype(numpy.float32, copy=True)

            axes = range(array.ndim) if self.axes is None else self.axes

            for axis in reversed(axes):
                lprint(f"Correcting along axis: {axis}")
                array = self._permutate(array, axis=axis)
                shifts = self._shifts[axis]
                array = _shift_transform(
                    array, shifts, pad=False, crop=self.crop, pad_mode=''
                )
                array = self._depermutate(array, axis=axis)

            # cast back to original dtype:
            array = array.astype(self._original_dtype, copy=False)

            return array

    def _permutate(self, array, axis: int):
        permutation = self._get_permutation(array, axis=axis)
        array = numpy.transpose(array, axes=permutation)
        return array

    def _depermutate(self, array, axis: int):
        permutation = self._get_permutation(array, axis=axis, inverse=True)
        array = numpy.transpose(array, axes=permutation)
        return array

    def _get_permutation(self, array, axis: int, inverse=False):
        permutation = (axis,) + tuple(a for a in range(array.ndim) if a != axis)
        if inverse:
            permutation = numpy.argsort(permutation)
        return permutation


def _shift_transform(array, shifts, pad, crop, pad_mode='wrap'):
    """ """

    min_shift = abs(numpy.min(shifts, axis=0))
    max_shift = abs(numpy.max(shifts, axis=0))
    lprint(f"min_shift: {min_shift}, max_shift: {max_shift}")

    if pad:
        padding = tuple((mi, ma) for mi, ma in zip(min_shift, max_shift))
        lprint(f"Padding: {padding}")

        value = 0
        value = array.mean() if pad_mode == 'mean_constant' else value
        value = array.min() if pad_mode == 'min_constant' else value
        value = array.max() if pad_mode == 'max_constant' else value
        kwargs = {'constant_values': value} if 'constant' in pad_mode else {}

        pad_mode = 'constant' if 'constant' in pad_mode else pad_mode
        array = numpy.pad(array, pad_width=((0, 0),) + padding, mode=pad_mode, **kwargs)

    for ti, shift in enumerate(shifts):
        lprint(f"Motion correcting {ti} by {shift}")
        array[ti, ...] = numpy.roll(
            array[ti, ...], shift=shift, axis=tuple(range(0, array.ndim - 1))
        )

    if crop:
        crop_slice = (slice(None),) + tuple(
            slice(ma, s - mi)
            for mi, ma, s in zip(min_shift, max_shift, array.shape[1:])
        )
        array = array[crop_slice]

    return array


def _measure_shifts(
    array,
    reference_index: Optional[int] = None,
    center: bool = False,
    max_pixel_shift: Optional[int] = None,
    mode: str = 'com',
    sigma: float = 7,
):
    shifts = []
    correlations = []

    reference_index = len(array) // 2 if reference_index is None else reference_index
    max_pixel_shift = (
        max(array.shape[1:]) // 3 if max_pixel_shift is None else max_pixel_shift
    )

    for i in range(0, len(array)):
        image = array[i]
        if reference_index >= 0:
            reference_image = array[reference_index]
        elif reference_index < 0:
            reference_image = array[max(0, i + reference_index)]

        shift, correlation = _find_shift(
            image,
            reference_image,
            max_pixel_shift=max_pixel_shift,
            mode=mode,
            sigma=sigma,
        )
        shifts.append(shift)
        correlations.append(correlation)
        lprint(f"Measured shift of {shift} for image {i}.")

    shifts = numpy.array(shifts)
    if reference_index < 0:
        shifts = numpy.cumsum(shifts, axis=0)

    # correlations = numpy.stack(correlations)
    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(array, name='array')
    #     viewer.add_image(correlations, name='correlations')

    # Center shifts:
    if center:
        mean_shift = numpy.round(numpy.mean(shifts, axis=0))
        mean_shift = mean_shift.astype(numpy.int)
        shifts = shifts - mean_shift
    else:
        mean_shift = numpy.array((0,) * (array.ndim - 1))

    # Convert to int:
    shifts = numpy.round(shifts).astype(numpy.int)

    return shifts, mean_shift


def _find_shift(a, b, max_pixel_shift: int = 64, mode: str = 'com', sigma: float = 7):
    # Basic idea: We just need to low-pass filter the heck of it, and it works.
    lprint(f"max_pixel_shift: {max_pixel_shift}, mode: {mode}, sigma: {sigma}")

    # First we blur the input images:
    a = _fast_denoise(a, sigma=sigma)
    b = _fast_denoise(b, sigma=sigma)

    # We compute the phas correlation:
    raw_correlation = _phase_correlation(a, b)

    # We denoise the correlogram itself again:
    correlation = _fast_denoise(raw_correlation, sigma=sigma)
    # correlation = raw_correlation

    # We estimate the noise floor of the correlation:
    empty_region = correlation.copy()
    empty_region_slice = tuple(
        slice(min(s // 2 - 1, max_pixel_shift), -min(s // 2 - 1, max_pixel_shift))
        for s in correlation.shape
    )
    empty_region = empty_region[empty_region_slice]
    noise_floor_level = numpy.percentile(empty_region, q=99.9)

    # we use that floor to clip anything below:
    correlation = correlation.clip(noise_floor_level, math.inf) - noise_floor_level

    # We roll the array and crop it to restrict ourself to the search region:
    correlation = numpy.roll(
        correlation, shift=max_pixel_shift, axis=tuple(range(a.ndim))
    )
    correlation = correlation[(slice(0, 2 * max_pixel_shift),) * a.ndim]

    if mode == 'gfit':

        # This looks fancy and shit but is just bad.

        def gaussian(x, sx=0, sy=0, b=1, a=1):
            return a * numpy.exp(-b * ((x[0] - sx) ** 2 + (x[1] - sy) ** 2))

        x = numpy.arange(0, 2 * max_pixel_shift, 1)
        y = numpy.arange(0, 2 * max_pixel_shift, 1)
        xx, yy = numpy.meshgrid(x, y)
        xdata = numpy.stack([xx, yy]).reshape(2, -1)
        ydata = correlation[xx, yy].reshape(-1)

        popt, pcov = curve_fit(
            gaussian,
            xdata,
            ydata,
            method='trf',
            p0=(0, 0, 1, 1),
            bounds=(
                [-max_pixel_shift, -max_pixel_shift, 0, 0],
                [+max_pixel_shift, +max_pixel_shift, numpy.inf, numpy.inf],
            ),
        )

        shift = popt[:2]

    elif mode == 'com':

        # This is simple, and works brilliantly, even with tons of noise.

        # We use the max as quickly computed proxy for the real center:
        rough_shift = numpy.unravel_index(
            numpy.argmax(correlation, axis=None), correlation.shape
        )

        # We crop further to facilitate center-of-mass estimation:
        fine_window_radius = 4 * sigma
        cropped_correlation_slice = tuple(
            slice(
                max(0, int(rs - fine_window_radius)),
                min(s - 1, int(rs + fine_window_radius)),
            )
            for rs, s in zip(rough_shift, correlation.shape)
        )
        lprint(f"Cropped correlation: {cropped_correlation_slice}")
        cropped_correlation = correlation[cropped_correlation_slice]

        # We compute the signed rough shift
        signed_rough_shift = numpy.array(rough_shift) - max_pixel_shift

        if numpy.all(cropped_correlation == 0):
            # No mass:
            signed_com_shift = numpy.zeros_like(signed_rough_shift)
        else:
            # We compute the center of mass:
            # We take the square to squash small values far from the maximum that are likely noisy...
            signed_com_shift = (
                numpy.array(scipy.ndimage.center_of_mass(cropped_correlation ** 2))
                - fine_window_radius
            )

        # The final shift is the sum of the rough sight plus the fine center of mass shift:
        shift = signed_rough_shift + signed_com_shift

    #
    shift = numpy.nan_to_num(shift)

    return shift, correlation


def _fast_denoise(array, sigma):
    denoised = gaussian_filter(array, sigma=sigma, mode='wrap')
    return denoised


def _phase_correlation(image, reference_image):
    G_a = scipy.fft.fftn(image, workers=-1)
    G_b = scipy.fft.fftn(reference_image, workers=-1)
    conj_b = numpy.ma.conjugate(G_b)
    R = G_a * conj_b
    R /= numpy.absolute(R)
    r = scipy.fft.ifftn(R, workers=-1).real
    return r



import numpy


# Test masking method ====================================================================
from aydin.nn.models.unet import UNetModel
from aydin.nn.util.receptive_field import bbox_idx, receptive_field_model


def test_receptive_field_masking():
    batch_norm = None
    supervised = False
    shiftconv = False  # True
    batch_size = 1
    num_lyrs = 2
    input_image = numpy.zeros((1, 512, 512, 1))
    input_shape = input_image.shape
    input_image[0, int(input_shape[1] / 2), int(input_shape[2] / 2), 0] = 1
    model = UNetModel(
        input_shape[1:],
        mini_batch_size=1,
        nb_unet_levels=num_lyrs,
        normalization=batch_norm,
        supervised=supervised,
        shiftconv=shiftconv,
        spacetime_ndim=2,
    )
    # Predict test 0-image
    if not shiftconv and not supervised:
        pred1 = model.predict(
            [input_image, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
        )
    else:
        pred1 = model.predict(input_image, batch_size=batch_size, verbose=1)

    # Calculate receptive field
    receptive_field_idx = numpy.array(bbox_idx(pred1))
    receptive_field_size = receptive_field_idx[:, -1] - receptive_field_idx[:, 0] + 1

    rf_size_test = receptive_field_model(model, verbose=True)
    assert rf_size_test == receptive_field_size[1:2]


# # Test shiftconv method ====================================================================
# def test_receptive_field_shiftconv():
#     batch_norm = None
#     supervised = False
#     shiftconv = True  # True
#     batch_size = 1
#     num_lyrs = 5
#     input_image = numpy.zeros((1, 512, 512, 1))
#     input_shape = input_image.shape
#     input_image[0, int(input_shape[1] / 2), int(input_shape[2] / 2), 0] = 1
#     model = UNet2DModel(
#         input_shape[1:],
#         rot_batch_size=1,
#         num_lyr=num_lyrs,
#         normalization=batch_norm,
#         supervised=supervised,
#         shiftconv=shiftconv,
#     )
#     # Predict test 0-image
#     if not shiftconv and not supervised:
#         pred1 = model.predict(
#             [input_image, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
#         )
#     else:
#         pred1 = model.predict(input_image, batch_size=batch_size, verbose=1)
#
#     # Calculate receptive field
#     receptive_field_idx = numpy.array(bbox_idx(pred1))
#     receptive_field_size = receptive_field_idx[:, -1] - receptive_field_idx[:, 0] + 1
#
#     rf_size_test = receptive_field_model(model, verbose=True)
#     assert rf_size_test == receptive_field_size[1:2]



from typing import Optional, Sequence, Tuple

import numpy
from numpy.random import rand
from scipy.ndimage import gaussian_filter

from aydin.features.groups.convolutional import ConvolutionalFeatures


class RandomFeatures(ConvolutionalFeatures):
    """
    Random Feature Group class
    """

    def __init__(self, size: int, num_features: Optional[int] = None):
        super().__init__(kernels=None)
        self.size = size
        self._num_features = num_features

        self.image = None
        self.exclude_center: Sequence[Tuple[int, ...]] = []

    def _ensure_random_kernels_available(self, ndim: int):
        # Ensures that the kernels are available for subsequent steps.
        # We can't construct the kernels until we know the dimension of the image
        if self.kernels is None or self.kernels[0].ndim != ndim:
            dct_kernels = []

            num_features = (
                self.size ** ndim if self._num_features is None else self._num_features
            )

            shape = tuple((self.size,) * ndim)

            for k in range(num_features):
                numpy.random.seed(k)
                kernel = numpy.where(rand(*shape) > 0.5, 1.0, 0.0)
                kernel = gaussian_filter(kernel, sigma=0.5)
                kernel /= kernel.sum()
                kernel = kernel.astype(numpy.float32, copy=False)

                # import napari
                # with napari.gui_qt():
                #      from napari import Viewer
                #      viewer = Viewer()
                #      viewer.add_image(kernel, name='kernel')

                dct_kernels.append(kernel)

            self.kernels = dct_kernels

    @property
    def receptive_field_radius(self) -> int:
        return self.size // 2

    def num_features(self, ndim: int) -> int:
        self._ensure_random_kernels_available(ndim)
        return super().num_features(ndim)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []
        self._ensure_random_kernels_available(image.ndim)
        super().prepare(image, excluded_voxels, **kwargs)



import numpy

from aydin.io.datasets import newyork
from aydin.it.transforms.range import RangeTransform


def demo_range(mode):
    image = newyork()

    rt = RangeTransform(mode=mode)

    preprocessed = rt.preprocess(image)
    postprocessed = rt.postprocess(preprocessed)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(preprocessed, name='preprocessed')
        viewer.add_image(postprocessed, name='postprocessed')

    assert postprocessed.shape == image.shape
    assert numpy.abs(postprocessed - image).mean() < 1e-8


demo_range("minmax")
demo_range("percentile")



import gc
import math
import multiprocessing
from os.path import join

import lightgbm
import numpy
from lightgbm import Booster

from aydin.regression.base import RegressorBase
from aydin.regression.gbm_utils.callbacks import early_stopping
from aydin.util.log.log import lsection, lprint


class LGBMRegressor(RegressorBase):
    """LightGBM Regressor."""

    def __init__(
        self,
        num_leaves: int = 127,
        max_num_estimators: int = int(1e4),
        max_bin: int = 512,
        learning_rate: float = 0.01,
        loss: str = 'l1',
        patience: int = 5,
        verbosity: int = -1,
        compute_load: float = 0.95,
        gpu_prediction: bool = False,
        compute_training_loss: bool = False,
    ):
        """Constructs a LightGBM regressor.

        Parameters
        ----------
        num_leaves
            Number of leaves
        max_num_estimators
            Maximum number of estimators
        max_bin
            Maximum number of allowed bins
        learning_rate
            Learning rate for the LightGBM model
        loss
            Type of loss to be used
        patience
            Number of rounds required for early stopping
        verbosity
        compute_load
            Allowed load on computational resources in percentage
        gpu_prediction : bool
        compute_training_loss : bool
            Flag to tell LightGBM whether to compute training loss or not

        """
        super().__init__()

        self.force_verbose_eval = False

        self.num_leaves = num_leaves
        self.max_num_estimators = max_num_estimators
        self.max_bin = max_bin
        self.learning_rate = learning_rate
        self.metric = loss
        self.early_stopping_rounds = patience
        self.verbosity = verbosity
        self.compute_load = compute_load
        self.gpu_prediction = gpu_prediction
        self.compute_training_loss = compute_training_loss  # This can be expensive

        self.opencl_predictor = None

        with lsection("LGBM Regressor"):
            lprint(f"learning rate: {self.learning_rate}")
            lprint(f"number of leaves: {self.num_leaves}")
            lprint(f"max bin: {self.max_bin}")
            lprint(f"n_estimators: {self.max_num_estimators}")
            lprint(f"patience: {self.early_stopping_rounds}")

    def _get_params(self, num_samples, dtype=numpy.float32):
        # min_data_in_leaf = 20 + int(0.01 * (num_samples / self.num_leaves))
        max_depth = max(3, int(int(math.log2(self.num_leaves))) - 1)
        max_bin = 256 if dtype == numpy.uint8 else self.max_bin

        lprint(f'learning_rate:  {self.learning_rate}')
        lprint(f'max_depth:  {max_depth}')
        lprint(f'num_leaves: {self.num_leaves}')
        lprint(f'max_bin:    {max_bin}')

        objective = self.metric
        if objective == 'l1':
            objective = 'regression_l1'
        elif objective == 'l2':
            objective = 'regression_l2'

        params = {
            "device": "cpu",
            "boosting_type": "gbdt",
            'objective': objective,
            "learning_rate": self.learning_rate,
            "num_leaves": self.num_leaves,
            "max_depth": max_depth,
            "max_bin": max_bin,
            "subsample_for_bin": 200000,
            "num_threads": max(1, int(self.compute_load * multiprocessing.cpu_count())),
            "metric": self.metric.lower(),
            'verbosity': -1,
            "bagging_freq": 1,
            "bagging_fraction": 0.8,
            "lambda_l1": 0.01,
            "lambda_l2": 0.01,
        }

        return params

    def _fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):

        with lsection("GBM regressor fitting:"):

            nb_data_points = y_train.shape[0]
            self.num_features = x_train.shape[-1]
            has_valid_dataset = x_valid is not None and y_valid is not None

            lprint(f"Number of data points: {nb_data_points}")
            if has_valid_dataset:
                lprint(f"Number of validation data points: {y_valid.shape[0]}")
            lprint(f"Number of features per data point: {self.num_features}")

            train_dataset = lightgbm.Dataset(x_train, y_train, silent=True)
            valid_dataset = (
                lightgbm.Dataset(x_valid, y_valid, silent=True)
                if has_valid_dataset
                else None
            )

            self.__epoch_counter = 0

            # We translate the it fgr callback into a lightGBM callback:
            # This avoids propagating annoying 'evaluation_result_list[0][2]'
            # throughout the codebase...
            def lgbm_callback(env):
                try:
                    val_loss = env.evaluation_result_list[0][2]
                except Exception as e:
                    val_loss = 0
                    lprint("Problem with getting loss from LightGBM 'env' in callback")
                    print(str(e))
                if regressor_callback:
                    regressor_callback(env.iteration, val_loss, env.model)
                else:
                    lprint(f"Epoch {self.__epoch_counter}: Validation loss: {val_loss}")
                    self.__epoch_counter += 1

            evals_result = {}

            verbose_eval = (lgbm_callback is None) or (self.force_verbose_eval)

            self.early_stopping_callback = early_stopping(
                self, self.early_stopping_rounds
            )

            with lsection("GBM regressor fitting now:"):
                model = lightgbm.train(
                    params=self._get_params(nb_data_points, dtype=x_train.dtype),
                    init_model=None,
                    train_set=train_dataset,
                    valid_sets=[valid_dataset, train_dataset]
                    if self.compute_training_loss
                    else valid_dataset,
                    early_stopping_rounds=None if has_valid_dataset else None,
                    num_boost_round=self.max_num_estimators,
                    callbacks=[lgbm_callback, self.early_stopping_callback]
                    if has_valid_dataset
                    else [lgbm_callback],
                    verbose_eval=verbose_eval,
                    evals_result=evals_result,
                )
                lprint("GBM fitting done.")

            del train_dataset
            del valid_dataset

            if has_valid_dataset:
                self.last_valid_loss = evals_result['valid_0'][self.metric][-1]

            if self.compute_training_loss:
                loss_history = {
                    'training': evals_result['training'][self.metric],
                    'validation': evals_result['valid_0'][self.metric],
                }
            else:
                loss_history = {'validation': evals_result['valid_0'][self.metric]}

            gc.collect()
            return _LGBMModel(model, self.gpu_prediction, loss_history)


class _LGBMModel:
    def __init__(self, model, gpu_prediction, loss_history):
        self.model = model
        self.gpu_prediction = gpu_prediction
        self.loss_history = loss_history

    def _save_internals(self, path: str):
        if self.model is not None:
            lgbm_model_file = join(path, 'lgbm_model.txt')
            self.model.save_model(lgbm_model_file)

    def _load_internals(self, path: str):
        lgbm_model_file = join(path, 'lgbm_model.txt')
        self.model = Booster(model_file=lgbm_model_file)

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['model']
        return state

    def predict(self, x):
        with lsection("GBM regressor prediction:"):

            lprint(f"Number of data points             : {x.shape[0]}")
            lprint(f"Number of features per data points: {x.shape[-1]}")

            lprint("GBM regressor predicting now...")
            if self.gpu_prediction:
                try:
                    lprint("Attempting OpenCL-based regression.")
                    from aydin.regression.gbm_utils.opencl_prediction import (
                        GBMOpenCLPrediction,
                    )

                    if self.opencl_predictor is None:
                        self.opencl_predictor = GBMOpenCLPrediction()

                    prediction = self.opencl_predictor.predict(
                        self.model, x, num_iteration=self.model.best_iteration
                    )

                    # We clear the OpenCL ressources:
                    del self.opencl_predictor
                    self.opencl_predictor = None

                    return prediction
                except Exception:
                    lprint(
                        "Failed OpenCL-based regression, doing CPU based prediction."
                    )

            prediction = self.model.predict(x, num_iteration=self.model.best_iteration)
            # LGBM is annoying, it spits out float64s
            prediction = prediction.astype(numpy.float32, copy=False)
            lprint("GBM regressor predicting done!")
            return prediction



import gc
import time
from typing import Optional
import numpy

from aydin.features.base import FeatureGeneratorBase
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.it.balancing.data_histogram_balancer import DataHistogramBalancer
from aydin.it.base import ImageTranslatorBase
from aydin.regression.base import RegressorBase
from aydin.regression.cb import CBRegressor
from aydin.util.array.nd import nd_split_slices
from aydin.util.log.log import lprint, lsection
from aydin.util.offcore.offcore import offcore_array


class ImageTranslatorFGR(ImageTranslatorBase):
    """
    Feature Generation & Regression (FGR) based Image TranslatorFGR Image Translator
    """

    feature_generator: FeatureGeneratorBase

    def __init__(
        self,
        feature_generator: FeatureGeneratorBase = None,
        regressor: RegressorBase = None,
        balance_training_data: bool = False,
        voxel_keep_ratio: float = 1,
        max_voxels_for_training: Optional[int] = None,
        favour_bright_pixels: bool = False,
        **kwargs,
    ):
        """Constructs a FGR image translator. FGR image translators use feature generation
        and regression learning to acheive image translation.

        Parameters
        ----------
        feature_generator : FeatureGeneratorBase
            Feature generator.
        regressor : RegressorBase
            Regressor.
        balance_training_data : bool
            Limits number training entries per target
            value histogram bin.
        voxel_keep_ratio : float
            Ratio of the voxels to keep for training.
        max_voxels_for_training : int, optional
            Maximum number of the voxels that can be
            used for training.
        favour_bright_pixels : bool
            Marks bright pixels more favourable for
            the data histogram balancer.
        kwargs : dict
            Keyword arguments.
        max_memory_usage_ratio : float
            Maximum allowed memory load.
        max_tiling_overhead : float
            Maximum allowed margin overhead during tiling.
        """
        super().__init__(**kwargs)

        self.voxel_keep_ratio = voxel_keep_ratio
        self.balance_training_data = balance_training_data
        self.favour_bright_pixels = favour_bright_pixels
        self.feature_generator = (
            StandardFeatureGenerator()
            if feature_generator is None
            else feature_generator
        )
        self.regressor = CBRegressor() if regressor is None else regressor

        self.max_voxels_for_training = (
            self.regressor.recommended_max_num_datapoints()
            if max_voxels_for_training is None
            else max_voxels_for_training
        )

        # It seems empirically that excluding the central feature incurs no cost in quality,
        # simply because for every scale the next scale already partly covers these pixels.
        self.exclude_center_feature = True

        # Option to exclude the center value during translation. Set to True by default.
        self.exclude_center_value_when_translating = False

        # Advanced private functionality:
        # This field gives the opportunity to specify which channels must be 'passed-through'
        # directly as features. This is not intended to be used directly by users.
        self._passthrough_channels = None

        with lsection("FGR image translator"):
            lprint(f"balance training data: {self.balance_training_data}")

    def save(self, path: str):
        """Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------
        frozen

        """
        with lsection(f"Saving 'fgr' image translator to {path}"):
            frozen = super().save(path)
            frozen += self.feature_generator.save(path) + '\n'
            frozen += self.regressor.save(path) + '\n'

        return frozen

    def _load_internals(self, path: str):
        with lsection(f"Loading 'fgr' image translator from {path}"):
            self.feature_generator = FeatureGeneratorBase.load(path)
            self.regressor = RegressorBase.load(path)

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['feature_generator']
        del state['regressor']
        return state

    def stop_training(self):
        """Stops currently running training within the instance by calling
        the corresponding `stop_fit()` method on the regressor.
        """
        self.regressor.stop_fit()

    def _estimate_memory_needed_and_available(self, image):
        with lsection(
            "Estimating the amount of memory needed to store feature arrays:"
        ):
            num_spatio_temp_dim = len(image.shape[2:])
            dummy_image = numpy.zeros(
                shape=(1, 1) + (4,) * num_spatio_temp_dim, dtype=numpy.float32
            )

            x = self._compute_features(
                dummy_image,
                exclude_center_feature=self.exclude_center_feature,
                exclude_center_value=False,
            )
            num_features = x.shape[-1]
            feature_dtype = x.dtype

            (
                memory_needed,
                memory_available,
            ) = super()._estimate_memory_needed_and_available(image)

            """Here, 1.3 correction factor is chosen to have a bigger safety margin as we use
            more memory while generating features(namely uniform features) than the amount
            of memory we need after feature computations."""
            memory_needed = max(
                memory_needed, 1.3 * num_features * image.size * feature_dtype.itemsize
            )

            return memory_needed, memory_available

    def _compute_features(
        self,
        image,
        exclude_center_feature,
        exclude_center_value,
        features_last=True,
        num_reserved_features=0,
        image_slice=None,
        whole_image_shape=None,
    ):
        """Internal function that computes features for a given image.

        :param image: image
        :param exclude_center_feature: exclude center feature
        :param exclude_center_value: exclude center value
        :return: returns flattened array of features
        """

        with lsection(f"Computing features for image of shape {image.shape}:"):
            excluded_voxels = (
                None
                if self.blind_spots is None
                else list(
                    [
                        coordinate
                        for coordinate in self.blind_spots
                        if coordinate != (0,) * (image.ndim - 2)
                    ]
                )
            )

            lprint(f"exclude_center_feature = {exclude_center_feature}")
            lprint(f"exclude_center_value   = {exclude_center_value}")
            lprint(f"excluded_voxels        = {excluded_voxels}")

            excluded_voxels = (
                None
                if self.blind_spots is None
                else list(
                    [
                        coordinate
                        for coordinate in self.blind_spots
                        if coordinate != (0,) * (image.ndim - 2)
                    ]
                )
            )

            # If this is a part of a larger image, we can figure out what are the offsets and scales for the spatial features:
            spatial_feature_scale = (
                None
                if whole_image_shape is None
                else tuple(1.0 / s for s in whole_image_shape[2:])
            )
            spatial_feature_offset = (
                None if image_slice is None else tuple(s.start for s in image_slice[2:])
            )
            lprint(f"spatial_feature_scale     = {spatial_feature_scale}")
            lprint(f"spatial_feature_offset    = {spatial_feature_offset}")

            features = self.feature_generator.compute(
                image,
                exclude_center_feature=exclude_center_feature,
                exclude_center_value=exclude_center_value,
                num_reserved_features=num_reserved_features,
                passthrough_channels=self._passthrough_channels,
                feature_last_dim=False,
                excluded_voxels=excluded_voxels,
                spatial_feature_offset=spatial_feature_offset,
                spatial_feature_scale=spatial_feature_scale,
            )

            num_features = features.shape[0]
            x = features.reshape(num_features, -1)

            if features_last:
                x = numpy.moveaxis(x, 0, -1)

            return x

    def _train(
        self, input_image, target_image, train_valid_ratio, callback_period, jinv
    ):
        with lsection(
            f"Training image translator from image of shape {input_image.shape} to image of shape {target_image.shape}:"
        ):

            self.prepare_monitoring_images()

            num_input_channels = input_image.shape[1]
            num_target_channels = target_image.shape[1]
            normalised_input_shape = input_image.shape

            # Deal with jinv parameter;
            if jinv is None:
                exclude_center_value = self.self_supervised
            elif type(jinv) is tuple:
                exclude_center_value = jinv
            else:
                exclude_center_value = jinv

            # Prepare the splitting of train from valid data as well as balancing and decimation...
            self._prepare_split_train_val(input_image, target_image)

            # Tilling strategy is determined here:
            tilling_strategy, margins = self._get_tilling_strategy_and_margins(
                input_image,
                max_voxels_per_tile=self.max_voxels_per_tile,
                min_margin=self.tile_min_margin,
                max_margin=self.tile_max_margin,
            )
            lprint(f"Tilling strategy (just batches): {tilling_strategy}")
            lprint(f"Margins for tiles: {margins} .")

            # tile slice objects with margins:
            tile_slices_margins = list(
                nd_split_slices(
                    normalised_input_shape, nb_slices=tilling_strategy, margins=margins
                )
            )

            # Number of tiles:
            number_of_tiles = len(tile_slices_margins)
            lprint(f"Number of tiles (slices): {number_of_tiles}")

            # We initialise the arrays:
            x_train, x_valid, y_train, y_valid = (None,) * 4

            for idx, slice_margin_tuple in enumerate(tile_slices_margins):
                with lsection(
                    f"Current tile: {idx}/{number_of_tiles}, slice: {slice_margin_tuple} "
                ):

                    # We first extract the tile image:
                    input_image_tile = input_image[slice_margin_tuple]
                    target_image_tile = target_image[slice_margin_tuple]

                    x_tile = self._compute_features(
                        input_image_tile,
                        exclude_center_feature=self.exclude_center_feature,
                        exclude_center_value=exclude_center_value,
                        num_reserved_features=0,
                        features_last=False,
                        image_slice=slice_margin_tuple,
                        whole_image_shape=input_image.shape,
                    )
                    y_tile = target_image_tile.reshape(num_target_channels, -1)

                    num_features_tile = x_tile.shape[0]
                    num_entries_tile = y_tile.shape[1]

                    lprint(
                        f"Number of entries: {num_entries_tile}, features: {num_features_tile}, input channels: {num_input_channels}, target channels: {num_target_channels}"
                    )

                    # We split this tile's data into train and valid sets:
                    (
                        x_train_tile,
                        x_valid_tile,
                        y_train_tile,
                        y_valid_tile,
                    ) = self._do_split_train_val(
                        num_target_channels, train_valid_ratio, x_tile, y_tile
                    )

                    # We get rid of x and y to free memory:
                    del x_tile, y_tile
                    gc.collect()

                    # We now put the feature dimension to the back:
                    x_train_tile = numpy.moveaxis(x_train_tile, 0, -1)
                    x_valid_tile = numpy.moveaxis(x_valid_tile, 0, -1)

                    if x_train is None or x_valid is None:
                        x_train, x_valid, y_train, y_valid = (
                            x_train_tile,
                            x_valid_tile,
                            y_train_tile,
                            y_valid_tile,
                        )
                    else:
                        x_train = numpy.append(x_train, x_train_tile, axis=0)
                        x_valid = numpy.append(x_valid, x_valid_tile, axis=0)
                        y_train = numpy.append(y_train, y_train_tile, axis=1)
                        y_valid = numpy.append(y_valid, y_valid_tile, axis=1)

            lprint("Training now...")
            self.regressor.fit(
                x_train=x_train,
                y_train=y_train,
                x_valid=x_valid,
                y_valid=y_valid,
                regressor_callback=self.get_callback() if self.monitor else None,
            )

            self.loss_history = self.regressor.loss_history

    def prepare_monitoring_images(self):
        # Compute features for monitoring images:
        if self.monitor is not None and self.monitor.monitoring_images is not None:
            # Normalise monitoring images:
            normalised_monitoring_images = [
                self.shape_normaliser.normalise(
                    monitoring_image, batch_dims=None, channel_dims=None
                )
                for monitoring_image in self.monitor.monitoring_images
            ]

            # compute features proper:
            monitoring_images_features = [
                self._compute_features(
                    monitoring_image,
                    exclude_center_feature=self.exclude_center_feature,
                    exclude_center_value=False,
                    features_last=True,
                )
                for monitoring_image in normalised_monitoring_images
            ]
        else:
            monitoring_images_features = None
        # We keep these features handy...
        self.monitoring_datasets = monitoring_images_features

    def get_callback(self):
        # Regressor callback:
        def regressor_callback(iteration, val_loss, model):

            if val_loss is None:
                return

            current_time_sec = time.time()

            # Correct for dtype range:
            if self.feature_generator.dtype == numpy.uint8:
                val_loss /= 255
            elif self.feature_generator.dtype == numpy.uint16:
                val_loss /= 255 * 255

            if current_time_sec > self.last_callback_time_sec + self.callback_period:

                if self.monitoring_datasets and self.monitor:
                    predicted_monitoring_datasets = [
                        self.regressor.predict(x_m, models_to_use=[model])
                        for x_m in self.monitoring_datasets
                    ]
                    inferred_images = [
                        y_m.reshape(image.shape)
                        for (image, y_m) in zip(
                            self.monitor.monitoring_images,
                            predicted_monitoring_datasets,
                        )
                    ]

                    # for image in inferred_images:
                    for index in range(len(inferred_images)):
                        (
                            inferred_images[index],
                            _,
                            _,
                        ) = self.shape_normaliser.shape_normalize(
                            inferred_images[index]
                        )

                    denormalised_inferred_images = [
                        self.target_shape_normaliser.denormalise(inferred_image)
                        for inferred_image in inferred_images
                    ]

                    self.monitor.variables = (
                        iteration,
                        val_loss,
                        denormalised_inferred_images,
                    )
                elif self.monitor:
                    self.monitor.variables = (iteration, val_loss, None)

                self.last_callback_time_sec = current_time_sec
            else:
                pass
                # print(f"Iteration={iteration} metric value: {eval_metric_value} ")

        return regressor_callback

    def _prepare_split_train_val(self, input_image, target_image):
        # the number of voxels:
        num_of_voxels = input_image.size
        lprint(
            f"Image has: {num_of_voxels} voxels, at most: {self.max_voxels_for_training} voxels will be used for training or validation."
        )
        # This is the ratio of pixels to keep:
        max_voxels_keep_ratio = float(self.max_voxels_for_training) / num_of_voxels
        effective_keep_ratio = min(self.voxel_keep_ratio, max_voxels_keep_ratio)
        lprint(
            f"Given train ratio is: {self.voxel_keep_ratio}, max_voxels induced keep-ratio is: {max_voxels_keep_ratio}"
        )

        # For small images it is not worth having any limit or balance anything:
        num_voxels_in_small_image = 5 * 1e6
        effective_keep_ratio = (
            1.0 if num_of_voxels < num_voxels_in_small_image else effective_keep_ratio
        )
        balance_training_data = (
            False
            if num_of_voxels < num_voxels_in_small_image
            else self.balance_training_data
        )

        lprint(
            f"Data histogram balancer is: {'active' if balance_training_data else 'inactive'}"
        )
        lprint(f"Effective keep-ratio is: {effective_keep_ratio}")
        lprint(
            f"Favouring bright pixels: {'yes' if self.favour_bright_pixels else 'no'}"
        )

        # We decide on a 'batch' length that will be used to shuffle, select and then copy the training data...
        num_of_voxels_per_stack = input_image[2:].size
        batch_length = (
            16
            if num_of_voxels_per_stack < 1e5
            else (
                32
                if num_of_voxels_per_stack < 1e6
                else (
                    128
                    if num_of_voxels_per_stack < 1e7
                    else (512 if num_of_voxels_per_stack < 1e8 else 4096)
                )
            )
        )
        lprint(f"Using contiguous batches of length: {batch_length} ")
        # We create a balancer common to all tiles:
        balancer = DataHistogramBalancer(
            balance=balance_training_data,
            keep_ratio=effective_keep_ratio,
            favour_bright_pixels=self.favour_bright_pixels,
        )
        # Calibration of the balancer is done on the entire image:
        lprint("Calibrating balancer...")
        balancer.calibrate(target_image.ravel(), batch_length)

        # Keep both balancer and batch length
        self.train_val_split_balancer = balancer
        self.train_val_split_batch_length = batch_length

    def _do_split_train_val(
        self, num_target_channels: int, train_valid_ratio: float, x, y
    ):
        with lsection(
            f"Splitting train and test sets (train_test_ratio={train_valid_ratio}) "
        ):
            balancer: DataHistogramBalancer = self.train_val_split_balancer
            batch_length: int = self.train_val_split_batch_length

            nb_features = x.shape[0]
            nb_entries = y.shape[1]

            nb_split_batches = max(nb_entries // batch_length, 64)
            lprint(
                f"Creating random indices for train/val split (nb_split_batches={nb_split_batches})"
            )

            nb_split_batches_valid = int(train_valid_ratio * nb_split_batches)
            nb_split_batches_train = nb_split_batches - nb_split_batches_valid
            is_train_array = numpy.full(nb_split_batches, False)
            is_train_array[nb_split_batches_valid:] = True
            lprint(f"train/valid bool array created (length={is_train_array.shape[0]})")

            lprint("Shuffling train/valid bool array...")
            numpy.random.shuffle(is_train_array)

            lprint("Calculating number of entries for train and validation...")
            nb_entries_per_split_batch = max(1, nb_entries // nb_split_batches)
            nb_entries_train = nb_split_batches_train * nb_entries_per_split_batch
            nb_entries_valid = nb_split_batches_valid * nb_entries_per_split_batch

            lprint(
                f"Number of entries for training: {nb_entries_train} = {nb_split_batches_train}*{nb_entries_per_split_batch}, validation: {nb_entries_valid} = {nb_split_batches_valid} * {nb_entries_per_split_batch}"
            )

            lprint("Allocating arrays...")
            x_train = offcore_array(
                shape=(nb_features, nb_entries_train),
                dtype=x.dtype,
                max_memory_usage_ratio=self.max_memory_usage_ratio,
            )
            y_train = offcore_array(
                shape=(num_target_channels, nb_entries_train),
                dtype=y.dtype,
                max_memory_usage_ratio=self.max_memory_usage_ratio,
            )
            x_valid = offcore_array(
                shape=(nb_features, nb_entries_valid),
                dtype=x.dtype,
                max_memory_usage_ratio=self.max_memory_usage_ratio,
            )
            y_valid = offcore_array(
                shape=(num_target_channels, nb_entries_valid),
                dtype=y.dtype,
                max_memory_usage_ratio=self.max_memory_usage_ratio,
            )

            with lsection("Copying data for training and validation sets..."):

                # We use a random permutation to avoid having the balancer drop only from the 'end' of the image
                permutation = numpy.random.permutation(nb_split_batches)

                i, jt, jv = 0, 0, 0
                dst_stop_train = 0
                dst_stop_valid = 0

                balancer.initialise(nb_split_batches)

                for is_train in numpy.nditer(is_train_array):
                    if i % (nb_split_batches // 64) == 0:
                        lprint(
                            f"Copying section [{i},{min(nb_split_batches, i + nb_split_batches // 64)}]"
                        )

                    permutated_i = permutation[i]
                    src_start = permutated_i * nb_entries_per_split_batch
                    src_stop = src_start + nb_entries_per_split_batch
                    i += 1

                    xsrc = x[:, src_start:src_stop]
                    ysrc = y[:, src_start:src_stop]

                    if balancer.add_entry(ysrc):
                        if is_train:
                            dst_start_train = jt * nb_entries_per_split_batch
                            dst_stop_train = (jt + 1) * nb_entries_per_split_batch

                            jt += 1

                            xdst = x_train[:, dst_start_train:dst_stop_train]
                            ydst = y_train[:, dst_start_train:dst_stop_train]

                            numpy.copyto(xdst, xsrc)
                            numpy.copyto(ydst, ysrc)

                        else:
                            dst_start_valid = jv * nb_entries_per_split_batch
                            dst_stop_valid = (jv + 1) * nb_entries_per_split_batch

                            jv += 1

                            xdst = x_valid[:, dst_start_valid:dst_stop_valid]
                            ydst = y_valid[:, dst_start_valid:dst_stop_valid]

                            numpy.copyto(xdst, xsrc)
                            numpy.copyto(ydst, ysrc)

                # Now we actually truncate out all the zeros at the end of the arrays:
                x_train = x_train[:, 0:dst_stop_train]
                y_train = y_train[:, 0:dst_stop_train]
                x_valid = x_valid[:, 0:dst_stop_valid]
                y_valid = y_valid[:, 0:dst_stop_valid]

                lprint(f"Histogram all    : {balancer.get_histogram_all_as_string()}")
                lprint(f"Histogram kept   : {balancer.get_histogram_kept_as_string()}")
                lprint(
                    f"Histogram dropped: {balancer.get_histogram_dropped_as_string()}"
                )
                lprint(
                    f"Number of entries kept: {balancer.total_kept()} out of {balancer.total_entries} total"
                )
                lprint(
                    f"Percentage of data kept: {100 * balancer.percentage_kept():.3f}% (train_data_ratio={balancer.keep_ratio}) "
                )
                if balancer.keep_ratio >= 1 and balancer.percentage_kept() < 1:
                    lprint(
                        "Note: balancer has dropped entries that fell on over-represented histogram bins"
                    )
        return x_train, x_valid, y_train, y_valid

    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        """Internal method that translates an input image on the basis of the trained model.

        :param input_image: input image
        :param batch_dims: batch dimensions
        :return:
        """
        shape = input_image.shape
        num_batches = shape[0]

        x = self._compute_features(
            input_image,
            exclude_center_feature=self.exclude_center_feature,
            exclude_center_value=self.exclude_center_value_when_translating,
            num_reserved_features=0,
            features_last=True,
            image_slice=image_slice,
            whole_image_shape=whole_image_shape,
        )

        with lsection(
            f"Predict from feature vector of dimension {x.shape} and dtype: {x.dtype}:"
        ):
            lprint("Predicting... ")
            # Predict using regressor:
            yp = self.regressor.predict(x)

            # We make sure that we have the result in float type, but make _sure_ to avoid copying data:
            if yp.dtype != numpy.float32 and yp.dtype != numpy.float64:
                yp = yp.astype(numpy.float32, copy=False)

            # We reshape the array:
            num_target_channels = yp.shape[0]
            translated_image_shape = (num_batches, num_target_channels) + shape[2:]
            lprint(f"Reshaping array to {translated_image_shape}... ")
            inferred_image = yp.reshape(translated_image_shape)

        return inferred_image



import multiprocessing

import numpy
from joblib import Parallel, delayed
from scipy.ndimage import uniform_filter

from aydin.util.array.nd import nd_split_slices, remove_margin_slice


def parallel_uniform_filter(
    image, size=3, output=None, mode="nearest", cval=0.0, origin=0
):
    # Save original image dtype:
    original_dtype = image.dtype

    # Scipy does not support float16 yet:
    dtype = numpy.float32 if original_dtype == numpy.float16 else original_dtype
    image = image.astype(dtype=dtype, copy=False)

    # Number of parallel jobs:
    num_jobs = max(1, int(0.95 * multiprocessing.cpu_count()))

    # TODO: adjust the number of jobs to the filter size

    # Longest axis:
    longest_axis_length = max(image.shape)

    # pick the longest axis for splitting:
    longest_axis = list(image.shape).index(longest_axis_length)

    # Normalise filter size to tuple:
    size = size if isinstance(size, tuple) else (size,) * image.ndim

    # If the filter size is too large, there is no point to split:
    filter_size_along_longest_axis = size[longest_axis]
    if (
        longest_axis_length // num_jobs + filter_size_along_longest_axis
        < longest_axis_length * 0.9  # we need to gain at least 10% speed!
    ):
        # No point in going parallel, we won't gain anything:
        output = uniform_filter(image, size=size, mode=mode, cval=cval, origin=origin)
    else:
        # configure splitting:
        nb_slices = [1] * image.ndim
        nb_slices[longest_axis] = min(num_jobs, image.shape[longest_axis])
        nb_slices = tuple(nb_slices)
        margins = (
            (size[longest_axis] // 2,) * image.ndim
            if isinstance(size, tuple)
            else (size // 2,) * image.ndim
        )

        # Obtain slice objects for splitting:
        slice_tuple_list = list(
            nd_split_slices(image.shape, nb_slices, do_shuffle=False)
        )
        slice_margin_tuple_list = list(
            nd_split_slices(image.shape, nb_slices, do_shuffle=False, margins=margins)
        )

        # Instantiates output array:
        output = numpy.empty_like(image)

        def _uniform_filter(slice_tuple, slice_margin_tuple):
            tile = image[slice_margin_tuple]

            output_tile = uniform_filter(
                tile, size=size, mode=mode, cval=cval, origin=origin
            )

            remove_margin_slice_tuple = remove_margin_slice(
                image.shape, slice_margin_tuple, slice_tuple
            )
            output_tile_without_margin = output_tile[remove_margin_slice_tuple]

            output[slice_tuple] = output_tile_without_margin

        slices = (
            (st, smt) for st, smt in zip(slice_tuple_list, slice_margin_tuple_list)
        )

        from joblib import parallel_backend

        with parallel_backend('threading', n_jobs=num_jobs):
            Parallel()(delayed(_uniform_filter)(st, smt) for st, smt in slices)

    output = output.astype(dtype=original_dtype, copy=False)

    return output



import sys


class RecursionLimit:
    """
    Recursion Limit

    Parameters
    ----------
    limit : int

    """

    def __init__(self, limit):
        self.limit = limit
        self.old_limit = sys.getrecursionlimit()

    def __enter__(self):
        sys.setrecursionlimit(self.limit)

    def __exit__(self, type, value, tb):
        sys.setrecursionlimit(self.old_limit)



from typing import Optional
import numpy

from aydin.it.base import ImageTranslatorBase
from aydin.it.transforms.padding import PaddingTransform
from aydin.util.log.log import lsection


class ImageTranslatorDeconvBase(ImageTranslatorBase):
    """Base class for Image Deconvolution"""

    def __init__(
        self,
        psf_kernel,
        *args,
        clip: bool = True,
        padding: Optional[int] = None,
        padding_mode=None,
        dtype=numpy.float32,
        max_voxels_per_tile: int = None,
        **kwargs,
    ):
        """Constructs a base deconvolution object

        Parameters
        ----------
        psf_kernel : numpy.typing.ArrayLike
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        args
        clip : bool
        padding : int, optional
        padding_mode
        dtype
        max_voxels_per_tile : int
        kwargs
        """
        super().__init__(*args, **kwargs)

        self.psf_kernel_numpy = psf_kernel
        self.clip = clip
        self.dtype = dtype
        self.__debug_allocation = False

        # default padding mode:
        if padding_mode is None:
            padding_mode = 'reflect'

        # remove existing padding:
        self.transforms_list = list(
            [t for t in self.transforms_list if not isinstance(t, PaddingTransform)]
        )
        padding = (
            max(max(7, s) for s in psf_kernel.shape) if padding is None else padding
        )
        # add correct padding:
        self.transforms_list.append(
            PaddingTransform(pad_width=padding, mode=padding_mode)
        )

        self.max_voxels_per_tile = max_voxels_per_tile

        self.tile_min_margin = max(psf_kernel.shape)

    def save(self, path: str):
        """Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------
        frozen

        """
        with lsection(f"Saving Lucy-Richardson image translator to {path}"):
            frozen = super().save(path)

        return frozen

    def _load_internals(self, path: str):
        """
        Method to load internals.

        Parameters
        ----------
        path : str
            path to load from

        """
        with lsection(f"Loading Lucy-Richardson image translator from {path}"):
            # no internals to load here...
            pass

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        # nothing to do here...
        return state

    def stop_training(self):
        """Stop training"""
        pass
        # we can't do that... for now...

    def _estimate_memory_needed_and_available(self, image):
        """
        Method to estimate needed and available memory amounts.

        Parameters
        ----------
        image : array_like

        Returns
        -------
        Tuple of memory_needed, memory_available

        """
        # By default there is no memory needed which means no constraints

        memory_needed, memory_available = super()._estimate_memory_needed_and_available(
            image
        )
        # TODO: this is a rough estimate, it is not clear how much is really needed...
        memory_needed = 6 * image.size * image.dtype.itemsize

        return memory_needed, memory_available

    def _debug_allocation(self, info):
        """
        Method to log used memory by CUDA with cupy backend.

        Parameters
        ----------
        info

        """
        pass



import numbers

import numpy

# from numpy.typing import ArrayLike
from scipy.ndimage import median_filter, gaussian_filter

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class HighpassTransform(ImageTransformBase):
    """Highpass Image Simplification

    For images with little noise, applying a high-pass filter can help denoise the image by removing some of the
    image complexity. The low-frequency parts of the image do not need to be denoised because sometimes the challenge
    is disentangling the (high-frequency) noise from the high-frequencies in the image. The scale parameter must be
    chosen with care. The lesser the noise, the smaller the value. Values around 1 work well but must be tuned
    depending on the image. If the scale parameter is too low, some noise might be left untouched. The best is to
    keep the parameter as low as possible while still achieving good denoising performance. It is also possible to
    apply median filtering when computing the low-pass image which helps reducing the impact of outlier voxel values,
    for example salt&pepper noise. Note: when median filtering is on, larger values of sigma (e.g. >= 1) are
    recommended, unless when the level of noise is very low in which case a sigma of 0 (no Gaussian blur) may be
    advantageous. To recover the original denoised image the filtering is undone during post-processing. Note: this
    is ideal for treating <a href='https://en.wikipedia.org/wiki/Colors_of_noise'>'blue' noise</a> that is
    characterised by a high-frequency support.
    """

    preprocess_description = (
        "Remove low spatial frequencies" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Add back low spatial frequencies" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        sigma: float = 1,
        median_filtering: bool = True,
        priority: float = 0.1,
        **kwargs,
    ):
        """
        Constructs a Highpass Transform

        Parameters
        ----------
        sigma : float
            Sigma value of the Gaussian filter used for high-pass filtering.
        median_filtering : bool
            Adds robustness of the filtering in the presence of outliers.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.sigma = sigma
        self.median_filtering = median_filtering
        self._low_pass_image = None
        self._original_dtype = None
        self._min = None
        self._max = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_low_pass_image']
        del state['_original_dtype']
        del state['_min']
        del state['_max']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (sigma={self.sigma}, median_filtering={self.median_filtering})'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Applies high-pass filter of sigma {self.sigma} {'and median filtering' if self.median_filtering else ''} to array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            # Remember min and max:
            self._min = array.min()
            self._max = array.max()

            # remember original dtype:
            self._original_dtype = array.dtype

            # Cast to float if needed:
            array = array.astype(numpy.float32, copy=False)

            # Low-pass filtering:
            self._low_pass_image = self._low_pass_filtering(array)

            # High-pass filtering:
            new_array = array - self._low_pass_image

            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Adds back low-pass frequencies to array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            array = array.astype(numpy.float32, copy=False)

            # Bring back the low frequencies:
            new_array = array + self._low_pass_image

            # If integer type we ensure to stay within original bounds:
            if issubclass(self._original_dtype.type, numbers.Integral):
                new_array = numpy.clip(new_array, self._min, self._max)

            # cast back to original dtype:
            new_array = new_array.astype(self._original_dtype, copy=False)

            # Free memory:
            self._low_pass_image = None

            return new_array

    def _low_pass_filtering(self, array):

        lprint(f"Sigma for high-pass filter is: {self.sigma}")

        # Median filtering if selected:
        if self.median_filtering:
            array = median_filter(array, size=3)

        # Compute mean:
        mean = numpy.mean(array).astype(numpy.float32)

        # Low-pass filtering:
        if self.sigma > 0:
            array = gaussian_filter(array, sigma=self.sigma) - mean

        return array


def _interpolation(image, x, y):
    out = numpy.interp(image.flat, x, y)
    return out.reshape(image.shape)



import importlib
import inspect
import os
import shutil
import numpy

from aydin import regression
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.it.base import ImageTranslatorBase
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.regression.cb import CBRegressor
from aydin.restoration.denoise.base import DenoiseRestorationBase
from aydin.util.log.log import lsection


if os.getenv("BUNDLED_AYDIN") == "1":
    from aydin.regression.lgbm import LGBMRegressor  # noqa: F401
    from aydin.regression.linear import LinearRegressor  # noqa: F401
    from aydin.regression.nn import NNRegressor  # noqa: F401
    from aydin.regression.rf import RFRegressor  # noqa: F401
    from aydin.regression.sv import SVRegressor  # noqa: F401


class Noise2SelfFGR(DenoiseRestorationBase):
    """
    Noise2Self image denoising "Feature Generation & Regression" (FGR)
    """

    def __init__(
        self,
        *,
        variant: str = 'fgr-cb',
        use_model=None,
        input_model_path=None,
        lower_level_args=None,
        it_transforms=None,
    ):
        """
        Parameters
        ----------
        variant : str
            Variant of N2S FGR denoising
        use_model : bool
            Flag to choose to train a new model or infer from a
            previously trained model. By default it is None.
        input_model_path : str
            Path to model that is desired to be used for inference.
            By default it is None.
        lower_level_args : args
            Additional 'low-level' arguments to be passed.
        it_transforms :
            Transforms to be applied.
        """

        super().__init__()
        self.lower_level_args = lower_level_args

        self.backend_it, self.backend_regressor = (
            variant.split("-") if variant is not None else ("fgr", "cb")
        )

        self.input_model_path = input_model_path
        self.use_model_flag = use_model
        self.model_folder_path = None

        self.it = None
        self.it_transforms = (
            [
                {"class": RangeTransform, "kwargs": {}},
                {"class": PaddingTransform, "kwargs": {}},
                {"class": VarianceStabilisationTransform, "kwargs": {}},
            ]
            if it_transforms is None
            else it_transforms
        )

        self.has_less_than_one_million_voxels = False
        self.has_less_than_one_trillion_voxels = True
        self.number_of_dims = -1

    @property
    def configurable_arguments(self):
        """Returns the configurable arguments that will be exposed
        on GUI and CLI.
        """
        arguments = {}

        # Feature Generator
        feature_generator = StandardFeatureGenerator
        fullargspec2 = inspect.getfullargspec(feature_generator.__init__)

        feature_generator_args = {
            "arguments": fullargspec2.args[4:],
            "defaults": fullargspec2.defaults[3:],
            "annotations": fullargspec2.annotations,
            "reference_class": feature_generator,
        }

        # IT FGR
        it = ImageTranslatorFGR

        fullargspec3 = inspect.getfullargspec(ImageTranslatorFGR.__init__)

        it_args = {
            "arguments": fullargspec3.args[3:],
            "defaults": fullargspec3.defaults[2:],
            "annotations": fullargspec3.annotations,
            "reference_class": it,
        }

        # Regressor
        regression_modules = DenoiseRestorationBase.get_implementations_in_a_module(
            regression
        )

        for module in regression_modules:
            regressor_args = self.get_class_implementation_kwonlyargs(
                regression, module, module.name.replace("_", "") + "Regressor"
            )

            arguments["Noise2SelfFGR-" + module.name] = {
                "feature_generator": feature_generator_args,
                "regressor": regressor_args,
                "it": it_args,
            }

        return arguments

    @property
    def implementations(self):
        """Returns the list of discovered implementations for given method."""
        return [
            "Noise2SelfFGR-" + x.name
            for x in self.get_implementations_in_a_module(regression)
        ]

    @property
    def implementations_description(self):
        fgr_description = Noise2SelfFGR.__doc__.strip()
        feature_generator_description = StandardFeatureGenerator.__doc__.strip()

        descriptions = []

        for module in self.get_implementations_in_a_module(regression):
            response = importlib.import_module(regression.__name__ + '.' + module.name)
            elem = [
                x
                for x in dir(response)
                if (module.name.replace("_", "") + "Regressor").lower() in x.lower()
            ][
                0
            ]  # class name

            elem_class = response.__getattribute__(elem)
            descriptions.append(
                fgr_description
                + ", uses "
                + feature_generator_description
                + " and uses "
                + elem_class.__doc__.replace("\n\n", "<br><br>")
            )

        return descriptions

    def stop_running(self):
        """Method to stop running N2S instance"""
        self.it.stop_training()

    def set_image_metrics(self, image_shape):
        """Sets several image metric parameters used internally.

        Parameters
        ----------
        image_shape : tuple

        """
        self.number_of_dims = len(image_shape)
        number_of_voxels = numpy.prod(numpy.array(image_shape))
        self.has_less_than_one_million_voxels = number_of_voxels < 1000000
        self.has_less_than_one_trillion_voxels = number_of_voxels < 1000000000000

    def get_generator(self):
        """Returns the corresponding generator instance for given selections.

        Returns
        -------
        generator : FeatureGeneratorBase

        """
        # print(self.lower_level_args)
        if self.lower_level_args is not None:
            generator = self.lower_level_args["feature_generator"]["class"](
                **self.lower_level_args["feature_generator"]["kwargs"]
            )
        else:
            generator = StandardFeatureGenerator()

        return generator

    def get_regressor(self):
        """Returns the corresponding regressor instance for given selections.

        Returns
        -------
        regressor : RegressorBase

        """

        if self.lower_level_args is not None:
            regressor = self.lower_level_args["regressor"]["class"](
                **self.lower_level_args["regressor"]["kwargs"]
            )
        else:
            regressor = CBRegressor()

        return regressor

    def get_translator(self, feature_generator, regressor):
        """Returns the corresponding translator instance for given selections.

        Parameters
        ----------
        feature_generator : FeatureGeneratorBase
        regressor : RegressorBase

        Returns
        -------
        it : ImageTranslatorBase

        """
        # Use a pre-saved model or train a new one from scratch and save it
        if self.use_model_flag:
            # Unarchive the model file and load its ImageTranslator object into self.it
            shutil.unpack_archive(
                self.input_model_path, os.path.dirname(self.input_model_path), "zip"
            )
            it = ImageTranslatorBase.load(self.input_model_path[:-4])
        else:
            it = ImageTranslatorFGR(
                feature_generator=feature_generator,
                regressor=regressor,
                **self.lower_level_args["it"]["kwargs"]
                if self.lower_level_args is not None
                else {},
                blind_spots='auto',  # TODO: ACS: please set this as default upstream
            )

        return it

    def add_transforms(self):
        if self.it_transforms is not None:
            for transform in self.it_transforms:
                transform_class = transform["class"]
                transform_kwargs = transform["kwargs"]
                self.it.add_transform(transform_class(**transform_kwargs))

    def train(
        self, noisy_image, *, batch_axes=None, chan_axes=None, image_path=None, **kwargs
    ):
        """Method to run training for Noise2Self FGR.

        Parameters
        ----------
        noisy_image : numpy.ndarray
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        image_path : str

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self train is starting..."):
            self.set_image_metrics(noisy_image.shape)

            self.it = self.get_translator(
                feature_generator=self.get_generator(), regressor=self.get_regressor()
            )

            self.add_transforms()

            # Train a new model
            self.it.train(
                noisy_image,
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                train_valid_ratio=kwargs['train_valid_ratio']
                if 'train_valid_ratio' in kwargs
                else 0.1,
                callback_period=kwargs['callback_period']
                if 'callback_period' in kwargs
                else 3,
                jinv=kwargs['jinv'] if 'jinv' in kwargs else None,
            )

            # Save the trained model
            # self.save_model(image_path)  # TODO:  fix the problems here

    def denoise(self, noisy_image, *, batch_axes=None, chan_axes=None, **kwargs):
        """Method to denoise an image with trained Noise2Self FGR.

        Parameters
        ----------
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        noisy_image : numpy.ndarray

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self denoise is starting..."):

            # Predict the resulting image
            response = self.it.translate(
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                tile_size=kwargs['tile_size'] if 'tile_size' in kwargs else None,
            )

            response = response.astype(noisy_image.dtype, copy=False)

            return response


def noise2self_fgr(noisy, *, batch_axes=None, chan_axes=None, variant=None):
    """Method to denoise an image with trained Noise2Self FGR.

    Parameters
    ----------
    noisy : numpy.ndarray
        Image to denoise
    batch_axes : array_like, optional
        Indices of batch axes.
    chan_axes : array_like, optional
        Indices of channel axes.
    variant : str
        Algorithm variant.

    Returns
    -------
    Denoised image : numpy.ndarray

    """
    # Run N2S and save the result
    n2s = Noise2SelfFGR(variant=variant)

    # Train
    n2s.train(noisy, batch_axes=batch_axes, chan_axes=chan_axes)

    # Denoise
    denoised = n2s.denoise(noisy, batch_axes=batch_axes, chan_axes=chan_axes)

    return denoised



import numpy


def check_nd(denoiser):
    shape = (17, 7, 13, 9)

    for ndim in range(1, 5):

        image = numpy.zeros(shape[:ndim])

        try:
            denoiser(image)
            assert True
        except Exception:
            assert False



import numpy
from numpy.random.mtrand import randint

from aydin.analysis.empirical_noise_model import (
    distill_noise_model,
    sample_noise_from_model,
)
from aydin.io.datasets import camera


def test_noise_model():
    clean_image = camera()

    noise = randint(0, 60, size=clean_image.size).reshape(*clean_image.shape)
    noisy_image = clean_image + noise

    noise_model = distill_noise_model(clean_image, noisy_image)

    noisy_image_sampled = sample_noise_from_model(clean_image, noise_model)

    diff = numpy.abs(noisy_image - noisy_image_sampled)

    assert (diff < 60).all()

    # import napari
    #
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(clean_image, name='clean_image')
    #     viewer.add_image(noisy_image, name='noisy_image')
    #     viewer.add_image(noisy_image_sampled, name='noisy_image_sampled')
    #     viewer.add_image(diff, name='diff')



from typing import Tuple


def default_patch_size(image, patch_size, odd: bool = True) -> Tuple[int, ...]:
    """
    Returns  a normalised patch size given an image amd a parity.
    Returns
    -------
    object
    """
    # Default patch sizes vary with image dimension:
    if patch_size is None:
        if image.ndim == 1:
            patch_size = 17 if odd else 16
        elif image.ndim == 2:
            patch_size = 8 if odd else 7
        elif image.ndim == 3:
            patch_size = 6 if odd else 5
        elif image.ndim == 4:
            patch_size = 4 if odd else 3

    # Normalise to tuple:
    if type(patch_size) is not tuple:
        patch_size = (patch_size,) * image.ndim

    # Ensure patch size is not larger than image:
    patch_size = tuple(
        min(ps, max(3, s // 2)) for ps, s in zip(patch_size, image.shape)
    )

    return patch_size



import numpy
import scipy
from numpy.random import normal

from aydin.analysis.blind_spot_analysis import auto_detect_blindspots
from aydin.io.datasets import camera, normalise, add_noise, rgbtest, examples_single


def test_blind_spot_analysis_simulated():
    image = camera()
    image = normalise(image.astype(numpy.float32, copy=False))
    image = add_noise(image)
    kernel = numpy.array([[0.2, 0.6, 0.2]])
    image = scipy.ndimage.convolve(image, kernel, mode='mirror')

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    assert blind_spots == [(0, -1), (0, 0), (0, 1)]


def test_blind_spot_analysis_2D_RGB():
    image = rgbtest().astype(numpy.float32)
    image += normal(0, 0.1, image.shape)

    blind_spots, _ = auto_detect_blindspots(image, channel_axes=(False, False, True))
    print(blind_spots)

    # TODO: this corelation is suspicious:
    assert blind_spots == [(-3, 0), (0, 0), (3, 0)]


def test_blind_spot_analysis_tribolium_2D():

    image = examples_single.myers_tribolium.get_array()
    image = image[20]

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    assert blind_spots == [(-1, 0), (0, 0), (1, 0)]


def test_blind_spot_analysis_tribolium_3D():
    image = examples_single.myers_tribolium.get_array()

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    assert blind_spots == [(0, -1, 0), (0, 0, 0), (0, 1, 0)]


def test_blind_spot_analysis_tribolium_3D_shallow():
    image = examples_single.myers_tribolium.get_array()
    image = image[10:-10]

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    assert blind_spots == [(0, -1, 0), (0, 0, 0), (0, 1, 0)]


def test_blind_spot_analysis_tribolium_3D_very_shallow():
    image = examples_single.myers_tribolium.get_array()
    image = image[0:4]

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    assert blind_spots == [(0, -1, 0), (0, 0, 0), (0, 1, 0)]


def test_blind_spot_analysis_tribolium_3D_super_shallow():
    image = examples_single.myers_tribolium.get_array()
    image = image[0]

    blind_spots, _ = auto_detect_blindspots(image)
    print(blind_spots)

    # NOTE: this might not be the right thing to do...
    assert blind_spots == [(-1, 0), (0, 0), (1, 0)]



import numbers
from math import ceil
from random import shuffle
import numpy
from numpy.lib.stride_tricks import as_strided


def nd_range(start, stop, dims):
    """
    nd_range

    Parameters
    ----------
    start : int
    stop : int
    dims : int

    """
    if not dims:
        yield ()
        return
    for outer in nd_range(start, stop, dims - 1):
        for inner in range(start, stop):
            yield outer + (inner,)


def nd_range_radii(radii):
    """
    nd_range_radii

    Parameters
    ----------
    radii : list

    """
    if not radii:
        yield ()
        return

    for outer in nd_range_radii(radii[:-1]):

        radius = radii[-1]
        start = -radius
        stop = +radius + 1

        for inner in range(start, stop):
            yield outer + (inner,)


def nd_range_bool_tuple(start, stop, dims):
    """
    nd_range_bool_tuple

    Parameters
    ----------
    start : int
    stop : int
    dims : int

    """
    if len(dims) == 0:
        yield ()
        return
    for outer in nd_range_bool_tuple(start, stop, dims[:-1]):
        if dims[-1]:
            for inner in range(start, stop):
                yield outer + (inner,)
        else:
            yield outer + ((start + stop) // 2,)


def nd_loop(stops):
    """
    nd_loop

    Parameters
    ----------
    stops : list

    """
    if not stops:
        yield ()
        return
    for outer in nd_loop(stops[:-1]):
        for inner in range(0, stops[-1]):
            yield outer + (inner,)


def nd_split_slices(array_shape, nb_slices, do_shuffle=False, margins=None):
    """
    nd_split_slices

    Parameters
    ----------
    array_shape : tuple
    nb_slices : list
    do_shuffle : tuple
    margins : tuple

    Returns
    -------

    """
    if not array_shape:
        yield ()
        return

    if margins is None:
        margins = (0,) * len(array_shape)

    dim_width = array_shape[-1]

    for outer in nd_split_slices(
        array_shape[:-1], nb_slices[:-1], do_shuffle=do_shuffle, margins=margins[:-1]
    ):

        n = nb_slices[-1]
        slice_width = int(ceil(dim_width / n))
        slice_margin = margins[-1]

        slice_start_range = list(range(0, dim_width, slice_width))

        if do_shuffle:
            shuffle(slice_start_range)

        for slice_start in slice_start_range:

            start = max(0, slice_start - slice_margin)
            stop = min(slice_start + slice_width + slice_margin, dim_width)
            yield outer + (slice(start, stop, 1),)


def remove_margin_slice(array_shape, slice_with_margin, slice_without_margin):
    """
    Remove_margin_slice

    Parameters
    ----------
    array_shape : tuple
    slice_with_margin : array_like
    slice_without_margin : array_like

    Returns
    -------
    sliced_tuple : tuple

    """
    slice_tuple = tuple(
        slice(max(0, v.start - u.start), min(v.stop - u.start, l), 1)
        for l, u, v in zip(array_shape, slice_with_margin, slice_without_margin)
    )
    return slice_tuple


def extract_tiles(arr, tile_size=8, extraction_step=1, flatten=False):
    """
    Extracts patches of any n-dimensional array in place using strides.
    Given an n-dimensional array it will return a 2n-dimensional array with
    the first n dimensions indexing patch position and the last n indexing
    the patch content. This operation is immediate (O(1)). A reshape
    performed on the first n dimensions will cause numpy to copy data, leading
    to a list of extracted patches.
    Read more in the :ref:`User Guide <image_feature_extraction>`.

    Parameters
    ----------
    arr : ndarray
        n-dimensional array of which patches are to be extracted
    tile_size : integer or tuple of length arr.ndim
        Indicates the shape of the patches to be extracted. If an
        integer is given, the shape will be a hypercube of
        sidelength given by its value.
    extraction_step : integer or tuple of length arr.ndim
        Indicates step size at which extraction shall be performed.
        If integer is given, then the step is uniform in all dimensions.
    Returns
    -------
    patches : strided ndarray
        2n-dimensional array indexing patches on first n dimensions and
        containing patches on the last n dimensions. These dimensions
        are fake, but this way no data is copied. A simple reshape invokes
        a copying operation to obtain a list of patches:
        result.reshape([-1] + list(patch_shape))
    """

    arr_ndim = arr.ndim

    if isinstance(tile_size, numbers.Number):
        tile_size = tuple([tile_size] * arr_ndim)
    if isinstance(extraction_step, numbers.Number):
        extraction_step = tuple([extraction_step] * arr_ndim)

    patch_strides = numpy.empty(arr.shape).strides

    slices = tuple(slice(None, None, st) for st in extraction_step)
    indexing_strides = numpy.empty(arr[slices].shape).strides

    patch_indices_shape = (
        (numpy.array(arr.shape) - numpy.array(tile_size))
        // numpy.array(extraction_step)
    ) + 1

    shape = tuple(list(patch_indices_shape) + list(tile_size))
    strides = tuple(list(indexing_strides) + list(patch_strides))

    patches = as_strided(arr, shape=shape, strides=strides)

    if flatten:
        patches = patches.reshape((-1,) + patches.shape[-arr.ndim :])

    return patches



# flake8: noqa
import numpy
import numpy as np
from skimage.data import camera

from aydin.io.datasets import normalise, add_noise, dots, lizard, pollen, newyork
from aydin.util.crop.demo.demo_rep_crop import demo_representative_crop
from aydin.util.crop.rep_crop import representative_crop
from aydin.util.log.log import Log


def test_representative_crop():
    newyork_image = newyork()
    demo_representative_crop(newyork_image, display=False)



import time

import numpy as np
from skimage.exposure import rescale_intensity

from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.it.cnn import ImageTranslatorCNN


def demo(image, max_epochs=10):
    batch_dims = (True, False, False, False)

    it = ImageTranslatorCNN(
        training_architecture='random',
        nb_unet_levels=2,
        batch_norm='instance',  # None,  #
        activation='ReLU',
        mask_size=3,
        # tile_size=128,
        # total_num_patches=10,
        max_epochs=max_epochs,
    )
    it.verbose = 1

    start = time.time()
    it.train(image, image, batch_axes=batch_dims)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised = it.translate(
        image, batch_axes=batch_dims, tile_size=100  # min(image.shape[1:-1])
    )
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")
    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(
            rescale_intensity(denoised, in_range='image', out_range=(0, 1)),
            name='denoised',
        )


image_path = examples_single.hyman_hela.get_path()
array, metadata = io.imread(image_path)
# array = array[0:10, 15:35, 130:167, 130:177]
array = array[:, :, 100:300, 100:300].astype(np.float32)
array = rescale_intensity(array, in_range='image', out_range=(0, 1))
demo(array)



import numpy

from aydin.util.numpy_memoize.memoize import memoize_last


def test_memoize():
    class Something(object):
        def __init__(self):
            self._private_val = 2
            self.ncalls = 0

        @memoize_last
        def f(self, x, k, *args, **kwargs):
            "This is a documented method."
            print('private val = ', self._private_val)  # Check access to attributes.
            self.ncalls += 1
            return numpy.random.random()

    something = Something()
    e = numpy.ones(5)
    print('First call:')
    val = something.f(e, 1, thingy='some stuff', z=3, y=-1, extra_arg='not important')
    print('return value = ', val)

    print('Second call (all same args):')

    val = something.f(e, 1, thingy='some stuff', z=3, y=-1, extra_arg='not important')
    print('return value = ', val)

    'Third call (first arg is different):'
    val = something.f(
        2 * e, 1, thingy='some stuff', z=3, y=-1, extra_arg='not important'
    )
    print('return value = ', val)

    print('')
    print('Number of evaluations: ', something.ncalls)
    assert something.ncalls == 2



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_harmonic import demo_harmonic
from aydin.it.classic_denoisers.harmonic import denoise_harmonic
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_harmonic():
    assert demo_harmonic(cropped_newyork(), display=False) >= 0.621 - 0.03


def test_harmonic_nd():
    check_nd(denoise_harmonic)



from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QTreeWidget,
    QTreeWidgetItem,
)

from aydin.gui._qt.custom_widgets.readmoreless_label import QReadMoreLessLabel
from aydin.io.io import FileMetadata


class DimensionsTab(QWidget):
    """
    Interpreting image dimensions <br><br> Images can have many dimensions: 2D, 3D, 4D, 3D+t... Some dimensions are
    'spatio-temporal' and the signal is expected to have a degree of continuity and correlation across these
    dimensions. Other dimensions are 'batch' dimensions, they just state that we have multiple images of the same
    kind and shape.

    <moreless> <br><br> Finally, some dimensions are 'channel' dimensions and carry vectorial information for each
    voxel of the image. In this tab, you can help aydin better denoise your images by telling it how to interpret the
    dimensions of your images. The choices made here will impact denoising speed and quality.

    In general, we don't recommend denoising multi-channel images unless the correlation between the channels is
    strong and expected. Denoising each channel separately reduces the chance of channel 'bleed-through'. If you
    expect continuity along certain dimensions do not select them as 'batch' dimensions.

    <split>
    In general, denoising leverages signal continuity and correlation and thus would benefit from as many 'spatio-temporal'
    dimensions as possible. However, the more spatio-temporal dimensions, the more intense the computation,
    and the longer the denoising. This can lead to better results but also can lead to 'blurring' effects if there is
    not enough continuity over these dimensions. For example, if the time interval between time steps is too long and
    consecutive images are too different from each other, it is often better to interpret that dimension as a batch
    dimension. In our experience, it is often better to trade speed for signal: if you can speed up imaging and
    augment signal continuity, even if the signal-to-noise ratio per image is worse, the fact that there are more
    images with different noise patterns but a highly continuous signal, will help achieve better denoising
    performance.
    """

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self._axes = None

        self.tab_layout = QVBoxLayout()
        self.tab_layout.setAlignment(Qt.AlignBottom)

        self.thumbnail_sliders_layout = QHBoxLayout()

        # Explanation text
        self.tab_layout.addWidget(QReadMoreLessLabel(self, self.__doc__))

        # Dimension list, batch and channel dimension selections
        self.dimensions_tree = QTreeWidget()
        self.dimensions_tree.setHeaderLabels(['Dimensions', 'T', 'Z', 'Y', 'X'])
        self.dimensions_tree.itemClicked.connect(self.onTreeClicked)
        self.dimensions_tree.setColumnWidth(0, 150)

        self.tab_layout.addWidget(self.dimensions_tree)

        self.setLayout(self.tab_layout)

    @property
    def axes(self):
        return self._axes

    @property
    def spatiotemporal_axes(self):
        return self._get_current_row_values(1)

    @property
    def batch_axes(self):
        return self._get_current_row_values(2)

    @property
    def channel_axes(self):
        return self._get_current_row_values(3)

    def _get_current_row_values(self, row_index):

        response = []
        root = self.dimensions_tree.invisibleRootItem()

        if root.childCount():
            column_count = root.child(0).columnCount() - 1
            for i in range(1, column_count + 1):
                item = root.child(row_index)
                response.append(bool(item.checkState(i)))

        return None if response == [] else response

    @property
    def dimensions(self):
        # TODO: change this to return a dict about dimensions
        response = []
        root = self.dimensions_tree.invisibleRootItem()

        if root.childCount():
            child_count = root.childCount()
            for i in range(child_count):
                item = root.child(i)
                response.append(item.text(0))

        return None if response == [] else response

    @dimensions.setter
    def dimensions(self, dimensions_metadata):
        if dimensions_metadata is None or dimensions_metadata == FileMetadata():
            self.remove_items_from_tree()
            self.parent.enable_disable_a_tab(self.__class__, False)
            return

        self.parent.enable_disable_a_tab(self.__class__, True)

        self.dimensions_tree.clear()

        for _ in range(len(dimensions_metadata.shape) + 1):
            self.dimensions_tree.header().showSection(_)

        for _ in range(
            len(dimensions_metadata.shape) + 1, self.dimensions_tree.header().count()
        ):
            self.dimensions_tree.header().hideSection(_)

        self.dimensions_tree.setHeaderLabels(
            ['Dimensions'] + [axis for axis in dimensions_metadata.axes]
        )
        self._axes = dimensions_metadata.axes

        self.size_row = QTreeWidgetItem(
            self.dimensions_tree,
            ['size'] + [str(axis_length) for axis_length in dimensions_metadata.shape],
        )

        self.spatiotemporal_row = QTreeWidgetItem(
            self.dimensions_tree,
            ['spatiotemporal'] + ['' for _ in dimensions_metadata.axes],
        )
        self.batch_row = QTreeWidgetItem(
            self.dimensions_tree, ['batch'] + ['' for _ in dimensions_metadata.axes]
        )
        self.channel_row = QTreeWidgetItem(
            self.dimensions_tree, ['channel'] + ['' for _ in dimensions_metadata.axes]
        )

        # Initialize the rows
        for ind, axis in enumerate(dimensions_metadata.batch_axes):
            self.spatiotemporal_row.setCheckState(ind + 1, Qt.Checked)

        for ind, axis in enumerate(dimensions_metadata.batch_axes):
            self.batch_row.setCheckState(ind + 1, Qt.Checked if axis else Qt.Unchecked)
            if axis:
                self.spatiotemporal_row.setCheckState(ind + 1, Qt.Unchecked)

        for ind, axis in enumerate(dimensions_metadata.channel_axes):
            self.channel_row.setCheckState(
                ind + 1, Qt.Checked if axis else Qt.Unchecked
            )
            if axis:
                self.spatiotemporal_row.setCheckState(ind + 1, Qt.Unchecked)

        self.handle_special_cases()

    @Slot(QTreeWidgetItem, int)
    def onTreeClicked(self, it, col):
        nb_spatiotemporal_dimensions = sum(int(x) for x in self.spatiotemporal_axes)
        clicked_row = ["spatiotemporal", "batch", "channel"].index(it.text(0))

        rows = [self.spatiotemporal_row, self.batch_row, self.channel_row]
        rows.pop(clicked_row)
        other_rows = rows

        # if clicked an item to check it
        if bool(it.checkState(col)):
            # clicked on spatiotemporal row
            if it.text(0) == "spatiotemporal":
                if nb_spatiotemporal_dimensions > 4:
                    it.setCheckState(col, Qt.Unchecked)
                else:
                    self.handle_exclusive_states(other_rows=other_rows, col_idx=col)
            # clicked on batch or channel row
            else:
                if (
                    nb_spatiotemporal_dimensions == 1
                    and self.spatiotemporal_row.checkState(col)
                ):
                    it.setCheckState(col, Qt.Unchecked)
                else:
                    self.handle_exclusive_states(other_rows=other_rows, col_idx=col)
        # if clicked an item to uncheck it
        else:
            # if other rows unchecked, check this row back
            for other_row in other_rows:
                if other_row.checkState(col) == Qt.Checked:
                    break
            else:
                it.setCheckState(col, Qt.Checked)

    def handle_special_cases(self):
        nb_spatiotemporal_dimensions = sum(int(x) for x in self.spatiotemporal_axes)

        if (
            nb_spatiotemporal_dimensions == 4
            and "T" in self.axes
            and not (
                self.batch_axes[self.axes.find("T")]
                or self.channel_axes[self.axes.find("T")]
            )
        ):
            self.batch_row.setCheckState(self.axes.find("T") + 1, Qt.Checked)
            self.spatiotemporal_row.setCheckState(self.axes.find("T") + 1, Qt.Unchecked)

    @staticmethod
    def handle_exclusive_states(other_rows, col_idx):
        for other_row in other_rows:
            if other_row.checkState(col_idx) == Qt.Checked:
                other_row.setCheckState(col_idx, Qt.Unchecked)

    def remove_items_from_tree(self):
        root = self.dimensions_tree.invisibleRootItem()
        child_count = root.childCount()
        for i in reversed(range(child_count)):
            item = root.child(i)

            (item.parent() or root).removeChild(item)

    def on_data_model_update(self):
        self.dimensions = (
            self.parent.data_model.images_to_denoise[0][2]
            if len(self.parent.data_model.images_to_denoise) == 1
            else None
        )



# flake8: noqa
from math import sin
import numpy

from aydin.util.optimizer.optimizer import Optimizer
from aydin.util.log.log import Log, lprint


def enable_log():
    # DO NOT REMOVE, FOR DEBUGGING PUPROSES!
    Log.enable_output = True
    Log.set_log_max_depth(5)
    pass


def demo_optimizer_mccormik():
    enable_log()

    # function to optimise:
    def mccormik(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -(sin(x + y) + (x - y) ** 2 - 1.5 * x + 2.5 * y + 1) - 1.9133

    x, v = Optimizer().optimize(
        mccormik, bounds=[(-1.5, 4.0), (-3.0, 4.0)], max_num_evaluations=512
    )

    print("\n Mccormik")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={-0.54719, -1.54719}")

    assert v > -1e-4

    assert numpy.allclose(x, [-0.54719, -1.54719], atol=1e-2)


def demo_optimizer_himmelblau():
    enable_log()

    # function to optimise:
    def himmelblau(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -((x ** 2 + y - 11) ** 2 + (x + y ** 2 - 7) ** 2)

    x, v = Optimizer().optimize(
        himmelblau, bounds=[(-5.0, 5.0), (-5.0, 5.0)], max_num_evaluations=512
    )

    print("\n Himmelblau")
    print(f"RESULT: x,v={x, v}")
    print(
        f"OPTIMUM:   x={3, 2}, or {-2.805, 3.131}, or {-3.779, -3.283}, or {3.584, -1.848}"
    )

    assert v > -1e-3


def demo_optimizer_beale():
    enable_log()

    # function to optimise:
    def beale(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -(
            (1.5 - x + x * y) ** 2
            + (2.25 - x + x * (y ** 2)) ** 2
            + (2.625 - x + x * y ** 3) ** 2
        )

    x, v = Optimizer().optimize(
        beale, bounds=[(-4.5, 4.5), (-4.5, 4.5)], max_num_evaluations=512
    )

    print("\n Beale")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={3, 0.5}")

    assert v > -1e-3

    # assert numpy.allclose(x, [3, 0.5], atol=1e-1)


def demo_optimizer_matyas():
    enable_log()

    # function to optimise:
    def matyas(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -(0.26 * (x ** 2 + y ** 2) - 0.48 * x * y)

    x, v = Optimizer().optimize(
        matyas, bounds=[(-10.0, 10.0), (-10.0, 10.0)], max_num_evaluations=512
    )

    print("\n Matyas")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={0, 0}")

    assert v > -1e-3

    assert numpy.allclose(x, [0, 0], atol=1e-2)


def demo_optimizer_goldsteinprice():
    enable_log()

    # function to optimise:
    def goldsteinprice(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return (
            1
            + (x + y + 1) ** 2
            * (19 - 14 * x + 3 * (x ** 2) - 14 * y + 6 * x * y + 3 * (y ** 2))
        ) * (
            (30 + (2 * x - 3 * y) ** 2)
            * (18 - 32 * x + 12 * (x ** 2) + 48 * y - 36 * x * y + 27 * (y ** 2))
        )

    x, v = Optimizer().optimize(
        goldsteinprice, bounds=[(-2.0, 2.0), (-2.0, 2.0)], max_num_evaluations=512
    )

    print("\n Goldstein & Price")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={0, -1}")

    assert v > -1e-3

    # assert numpy.allclose(x, [0, -1], atol=1e-1)


def demo_optimizer_rosenbrock():
    enable_log()

    # function to optimise:
    def rosenbrock(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -(100 * (y - x ** 2) ** 2 + (1 - x) ** 2)

    x, v = Optimizer().optimize(
        rosenbrock, bounds=[(-10.0, 10.0), (-10.0, 10.0)], max_num_evaluations=512
    )

    print("\n Rosenbrock")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM: x  ={1, 1}")

    assert v > -1e-1


#    assert numpy.allclose(x, [1, 1], atol=1e-1)


def demo_optimizer_sphere():
    enable_log()

    # function to optimise:
    def sphere(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -((x - 0.54321) ** 2 + (y + 0.332211) ** 2)

    x, v = Optimizer().optimize(
        sphere, bounds=[(-10.0, 10.0), (-10.0, 10.0)], max_num_evaluations=512
    )

    print("\n Sphere")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={0.54321, -0.332211}")

    assert v > -1e-3

    assert numpy.allclose(x, [0.54321, -0.332211], atol=1e-2)


def demo_optimizer_booth():
    enable_log()

    # function to optimise:
    def booth(x, y):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return -((x + 2 * y - 7) ** 2 + (2 * x + y - 5) ** 2)

    x, v = Optimizer().optimize(
        booth, bounds=[(-10.0, 10.0), (-10.0, 10.0)], max_num_evaluations=512
    )

    print("\n Booth")
    print(f"RESULT : x,v={x, v}")
    print(f"OPTIMUM:   x={1, 3}")

    assert v > -1e-1

    assert numpy.allclose(x, [1, 3], atol=1e-2)


def demo_optimizer():
    demo_optimizer_beale()
    demo_optimizer_mccormik()
    demo_optimizer_booth()
    demo_optimizer_sphere()
    demo_optimizer_rosenbrock()
    demo_optimizer_goldsteinprice()
    demo_optimizer_matyas()
    demo_optimizer_himmelblau()


if __name__ == "__main__":
    demo_optimizer()



import sys
from qtpy.QtCore import QThreadPool
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QMainWindow, QAction, QApplication, QStatusBar
import qdarkstyle

from aydin.gui.main_page import MainPage
from aydin.util.log.log import lprint


class App(QMainWindow):
    """GUI app"""

    def __init__(self, ver):
        super(QMainWindow, self).__init__()

        self.version = ver

        self.setWindowIcon(QIcon("resources/aydin_logo_grad_black.png"))

        self.threadpool = QThreadPool(self)

        self.title = "Aydin Studio - image denoising, but chill..."

        self.desktop = QApplication.desktop()
        self.screenRect = self.desktop.screenGeometry()
        height, width = self.screenRect.height(), self.screenRect.width()

        self.width = width // 3
        self.height = height // 2
        self.left = (width - self.width) // 2
        self.top = (height - self.height) // 2

        self.setWindowTitle(self.title)
        self.setGeometry(self.top, self.left, self.width, self.height)

        self.main_widget = MainPage(self, self.threadpool)
        self.setCentralWidget(self.main_widget)

        # Menu bar
        self.setupMenubar()

        # Status bar
        self.statusBar = QStatusBar(self)
        self.statusBar.showMessage(f"aydin, version: {ver}")
        self.setStatusBar(self.statusBar)

    def closeEvent(self, event):
        lprint("closeEvent of mainwindow is called")
        app = QApplication.instance()
        app.quit()

    def setupMenubar(self):
        """Method to populate menubar."""

        mainMenu = self.menuBar()
        mainMenu.setNativeMenuBar(False)
        fileMenu = mainMenu.addMenu(' &File')
        runMenu = mainMenu.addMenu(' &Run')
        helpMenu = mainMenu.addMenu(' &Help')

        # File Menu
        startPageButton = QAction('Add File(s)', self)
        startPageButton.setStatusTip('Add new files')
        startPageButton.triggered.connect(
            self.main_widget.tabs["File(s)"].openFileNamesDialog
        )
        fileMenu.addAction(startPageButton)

        exitButton = QAction(QIcon('exit24.png'), 'Exit', self)
        exitButton.setShortcut('Ctrl+Q')
        exitButton.setStatusTip('Exit application')
        exitButton.triggered.connect(self.close)
        fileMenu.addAction(exitButton)

        # Run Menu
        startButton = QAction('Start', self)
        startButton.setStatusTip('Start denoising')
        startButton.triggered.connect(
            self.main_widget.processing_job_runner.prep_and_run
        )
        runMenu.addAction(startButton)

        saveOptionsJSONButton = QAction('Save Options JSON', self)
        saveOptionsJSONButton.setStatusTip('Save options JSON')
        saveOptionsJSONButton.triggered.connect(
            lambda: self.main_widget.save_options_json()
        )
        runMenu.addAction(saveOptionsJSONButton)

        # loadOptionsJSONButton = QAction('Load Options JSON', self)
        # loadOptionsJSONButton.setStatusTip('Load options JSON')
        # # loadOptionsJSONButton.triggered.connect(
        # #     self.main_widget.tabs["File(s)"].openFileNamesDialog
        # # )
        # loadOptionsJSONButton.setEnabled(False)
        # runMenu.addAction(loadOptionsJSONButton)
        #
        # saveModelJSONButton = QAction('Save Model', self)
        # saveModelJSONButton.setStatusTip('Save the most-recent trained model')
        # saveModelJSONButton.setEnabled(False)
        # # saveModelJSONButton.triggered.connect(
        # #     self.main_widget.tabs["File(s)"].openFileNamesDialog
        # # )
        # runMenu.addAction(saveModelJSONButton)

        # Help Menu
        versionButton = QAction("ver" + self.version, self)
        helpMenu.addAction(versionButton)


def run(ver):
    """Method to run GUI

    Parameters
    ----------
    ver : str
        string of aydin version number

    """
    app = QApplication(sys.argv)
    app.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
    ex = App(ver)
    ex.show()
    sys.exit(app.exec())



# flake8: noqa
import numpy

from aydin.io.datasets import newyork
from aydin.util.patch_transform.patch_transform import (
    extract_patches_nd,
    reconstruct_from_nd_patches,
)


def test_patch_transform():
    _round_trip(7, display=False)
    _round_trip(8, display=False)


def _round_trip(patch_size: int, display: bool):
    # Image:
    image = newyork()
    # First we apply the patch transform:
    patches = extract_patches_nd(image, patch_size=patch_size)
    # Transform back from patches to image:
    reconstructed_image = reconstruct_from_nd_patches(patches, image.shape)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(patches, name='patches')
            viewer.add_image(reconstructed_image, name='reconstructed_image')
        assert numpy.allclose(image, reconstructed_image, atol=1)



# flake8: noqa
import time

import napari
import numpy
import numpy as np
from aydin.features.fast.fast_features import FastFeatureGenerator
from aydin.it.monitor import Monitor
from skimage.data import camera
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.util import random_noise

from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.nn import NNRegressor


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32, copy=False), in_range='image', out_range=(0, 1)
    )


def demo(regressor):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    image = camera().astype(np.float32)
    image = n(image)

    intensity = 5
    np.random.seed(0)
    noisy = np.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode='gaussian', var=0.01, seed=0)
    noisy = noisy.astype(np.float32)

    with napari.gui_qt():
        viewer = napari.Viewer()

        size = 128
        monitoring_image = noisy[
            256 - size // 2 : 256 + size // 2, 256 - size // 2 : 256 + size // 2
        ]

        def callback(arg):
            # print(arg)
            iteration, val_loss, image = arg
            print(
                f"********CALLBACK******** --> Iteration: {iteration} metric: {val_loss}"
            )
            # print(f"images: {str(images)}")
            # print("image: ", image[0])
            if image[0] is not None:
                print(
                    f"********CALLBACK******** --> Image: shape={image[0].shape} dtype={image[0].dtype}"
                )
                viewer.add_image(
                    rescale_intensity(image[0], in_range='image', out_range=(0, 1)),
                    name=f'noisy{iteration}',
                )

        generator = FastFeatureGenerator()

        monitor = Monitor(
            monitoring_callbacks=[callback], monitoring_images=[monitoring_image]
        )

        it = ImageTranslatorFGR(
            feature_generator=generator, regressor=regressor, monitor=monitor
        )

        start = time.time()

        it.train(noisy, noisy)

        stop = time.time()
        print(f"Training: elapsed time:  {stop - start} ")

        denoised = it.translate(noisy)
        denoised = rescale_intensity(denoised, in_range='image', out_range=(0, 1))

        print("noisy", psnr(image, noisy), ssim(noisy, image))
        print("denoised", psnr(image, denoised), ssim(denoised, image))

        viewer.add_image(n(image), name='image')
        viewer.add_image(n(noisy), name='noisy')
        viewer.add_image(n(denoised), name='denoised')


demo(NNRegressor(max_epochs=10))
demo(LGBMRegressor(max_num_estimators=400))



from typing import Union, Optional

from aydin.features.groups.convolutional import ConvolutionalFeatures
from aydin.features.groups.extract_kernels import extract_kernels


class LearnedConvolutionalFeatures(ConvolutionalFeatures):
    """
    Learned Convolutions Feature Group class
    """

    def __init__(
        self,
        size: int,
        num_kernels: Optional[int],
        num_patches: Union[int, float] = 1e5,
    ):
        super().__init__(kernels=None)
        self.size = size
        self.num_kernels = num_kernels
        self.num_patches = num_patches

        self.image = None
        self.exclude_center: bool = False

    @property
    def receptive_field_radius(self) -> int:
        return self.size // 2

    def num_features(self, ndim: int) -> int:
        return self.num_kernels

    def learn(self, image):
        self.kernels = extract_kernels(
            image,
            size=self.size,
            num_kernels=self.num_kernels,
            num_patches=self.num_patches,
        )

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        self.image = image

        super().prepare(image, excluded_voxels, **kwargs)



import math
from qtpy.QtWidgets import QCheckBox

from aydin.gui.tabs.qt.base_cropping import BaseCroppingTab
from aydin.util.crop.rep_crop import representative_crop


class TrainingCroppingTab(BaseCroppingTab):
    """
    Cropping Image for training & auto-tuning

    Use the sliders to select a region of the image to define the cropping region for training or auto-tuning. Aydin
    automatically suggests a crop based on the image content.

    <moreless>

    It is often not advised to use the entire image for training or auto-tuning. You may want to focus the attention
    of the algorithm to a specific region, exclude excessive dark background. or simply reduce the time required for
    training or auto-tuning. In the case of neural networks (CNN, NN) and dictionary based methods more data does
    help, methods using gradient boosting less so (N2S-FGR-cb/lgbm), and simpler methods such as low-pass filtering (
    Butterworth, Gaussian, ...) need very data (but ideally highly contrasted and detail rich regions) to work.

    <split>
    """

    def __init__(self, parent):
        super().__init__(parent)

        self.use_same_crop_checkbox = QCheckBox("Use same cropping for denoising")
        self.use_same_crop_checkbox.toggled.connect(
            lambda: parent.handle_use_same_crop_state_changed()
        )

        self.explanation_layout.addStretch()
        self.explanation_layout.addWidget(self.use_same_crop_checkbox, 10)

    @property
    def images(self):
        return super().images

    @images.setter
    def images(self, images):
        super(TrainingCroppingTab, self.__class__).images.fset(self, images)

        if len(images) == 1:
            image = images[0][1]
            response = representative_crop(
                image,
                mode='contrast' if image.size > 1_000_000 else 'sobelmin',
                crop_size=500_000,
                fast_mode=image.size > 2_000_000,
                fast_mode_num_crops=1024,
                return_slice=True,
            )

            if type(response) == tuple:
                best_slice = response[1]

                # TODO: extend this implementation to handle all chosen spatial-temporal dimensions
                y_slice = best_slice[images[0][2].axes.find("Y")]
                x_slice = best_slice[images[0][2].axes.find("X")]
                self.y_crop_slider.slider.setValues((y_slice.start, y_slice.stop))
                self.x_crop_slider.slider.setValues((x_slice.start, x_slice.stop))

    def update_summary(self):

        super().update_summary()

        if math.prod(self.cropped_image[0].shape) > 50_000_000:
            self.size_warning_label.setText(
                "Training might take a long time, please consider to reduce the size of crop as much as possible"
            )
            self.size_warning_label.setStyleSheet(
                "QLabel {font-weight: bold; color: red;}"
            )
        else:
            self.size_warning_label.setText("")



from m2cgen import ast
from m2cgen.assemblers import utils
from m2cgen.assemblers.boosting import (
    BaseTreeBoostingAssembler,
    LEAVES_CUTOFF_THRESHOLD,
)


class LightGBMModelAssembler(BaseTreeBoostingAssembler):

    classifier_names = {"LGBMClassifier"}

    def __init__(
        self, model, leaves_cutoff_threshold=LEAVES_CUTOFF_THRESHOLD, tree_slice=None
    ):
        model_dump = model.dump_model()
        if tree_slice is None:
            trees = [m["tree_structure"] for m in model_dump["tree_info"]]
        else:
            trees = [m["tree_structure"] for m in model_dump["tree_info"][tree_slice]]
            # [model_dump["tree_info"][tree_slice]["tree_structure"]]

        self.n_iter = len(trees) // model_dump["num_tree_per_iteration"]
        self.average_output = model_dump.get("average_output", False)

        super().__init__(model, trees, leaves_cutoff_threshold=leaves_cutoff_threshold)

    def _final_transform(self, ast_to_transform):
        if self.average_output:
            coef = 1 / self.n_iter
            return utils.apply_bin_op(
                ast_to_transform, ast.NumVal(coef), ast.BinNumOpType.MUL
            )
        else:
            return super()._final_transform(ast_to_transform)

    def _assemble_tree(self, tree):
        if "leaf_value" in tree:
            return ast.NumVal(tree["leaf_value"])

        threshold = ast.NumVal(tree["threshold"])
        feature_ref = ast.FeatureRef(tree["split_feature"])

        op = ast.CompOpType.from_str_op(tree["decision_type"])
        assert op == ast.CompOpType.LTE, "Unexpected comparison op"

        # Make sure that if the "default_left" is true the left tree branch
        # ends up in the "else" branch of the ast.IfExpr.
        if tree["default_left"]:
            op = ast.CompOpType.GT
            true_child = tree["right_child"]
            false_child = tree["left_child"]
        else:
            true_child = tree["left_child"]
            false_child = tree["right_child"]

        return ast.IfExpr(
            ast.CompExpr(feature_ref, threshold, op),
            self._assemble_tree(true_child),
            self._assemble_tree(false_child),
        )



import numpy
from skimage.data import camera
from skimage.exposure import rescale_intensity
from scipy.stats import entropy

from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.nn.util.data_util import random_sample_patches


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


# Test with garder 3D image
def test_random_sample_patch_3D():
    image_path = examples_single.hyman_hela.get_path()
    image0, metadata = io.imread(image_path)
    print(image0.shape)
    image0 = n(image0.squeeze()[0:32, :, 200:300, 200:300])

    image0 = numpy.expand_dims(image0[1:2], -1)
    tile_size = (32, 32, 32)
    num_tile = 100
    adoption_rate = 0.2
    input_data = random_sample_patches(image0, tile_size, num_tile, adoption_rate)

    # Extract patched images
    img_patch = []
    for i in input_data:
        img_patch.append(image0[i])
    img_patch = numpy.vstack(img_patch)
    # Entropy of the whole image
    hist, _ = numpy.histogram(image0, range=(0, 1), bins=255, density=True)
    entropy_whole = entropy(hist)

    # Entropy of sampled areas.
    hist, _ = numpy.histogram(img_patch, range=(0, 1), bins=255, density=True)
    entropy_smpl = entropy(hist)

    ent_ratio = entropy_smpl / entropy_whole

    assert ent_ratio >= 1.0


# Test with Cameraman
def test_random_sample_patch_2D():
    image0 = camera().astype(numpy.float32)
    image0 = numpy.expand_dims(numpy.expand_dims(n(image0), -1), 0)
    patch_size = (64, 64)
    num_patch = 500
    adoption_rate = 0.5
    input_data = random_sample_patches(image0, patch_size, num_patch, adoption_rate)

    # Extract patched images
    img_patch = []
    for i in input_data:
        img_patch.append(image0[i])
    img_patch = numpy.vstack(img_patch)
    # Entropy of the whole image
    hist, _ = numpy.histogram(image0, range=(0, 1), bins=255, density=True)
    entropy_whole = entropy(hist)

    # Entropy of sampled areas.
    hist, _ = numpy.histogram(img_patch, range=(0, 1), bins=255, density=True)
    entropy_smpl = entropy(hist)

    ent_ratio = entropy_smpl / entropy_whole

    assert ent_ratio >= 1.01


def test_random_sample_patch_2D_1patch():
    """
    Test whether random_sample_patches generates 1 image when patch size is same as input size.
    """
    image0 = camera().astype(numpy.float32)
    image0 = numpy.expand_dims(numpy.expand_dims(n(image0), -1), 0)
    patch_size = (512, 512)
    num_patch = 500
    adoption_rate = 0.5
    input_data = random_sample_patches(image0, patch_size, num_patch, adoption_rate)

    # Extract patched images
    img_patch = []
    for i in input_data:
        img_patch.append(image0[i])
    img_patch = numpy.vstack(img_patch)

    assert img_patch.shape == image0.shape



from tensorflow.python.keras.api.keras import (
    initializers,
    regularizers,
    constraints,
    backend,
)
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.engine.input_spec import InputSpec


class InstanceNormalization(Layer):
    """
    Instance normalization layer.

    Normalize the activations of the previous layer at each step,
    i.e. applies a transformation that maintains the mean activation
    close to 0 and the activation standard deviation close to 1.

    Parameters
    ----------
    axis : int
        the axis that should be normalized
        (typically the features axis).
        For instance, after a `Conv2D` layer with
        `data_format="channels_first"`,
        set `axis=1` in `InstanceNormalization`.
        Setting `axis=None` will normalize all values in each
        instance of the batch.
        Axis 0 is the batch dimension. `axis` cannot be set to 0 to avoid errors.
    epsilon : float
        Small float added to variance to avoid dividing by zero.
    center : bool
        If True, add offset of `beta` to normalized tensor.
        If False, `beta` is ignored.
    scale : bool
        If True, multiply by `gamma`.
        If False, `gamma` is not used.
        When the next layer is linear (also e.g. `nn.relu`),
        this can be disabled since the scaling
        will be done by the next layer.
    beta_initializer : string
        Initializer for the beta weight.
    gamma_initializer : string
        Initializer for the gamma weight.
    beta_regularizer : string
        Optional regularizer for the beta weight.
    gamma_regularizer : string
        Optional regularizer for the gamma weight.
    beta_constraint
    gamma_constraint
    kwargs

    Notes
    -----
    Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a Sequential model.

    Output shape
        Same shape as input.

    References
        - [Layer Normalization](https://arxiv.org/abs/1607.06450)
        - [Instance Normalization: The Missing Ingredient for Fast Stylization](
        https://arxiv.org/abs/1607.08022)

    """

    def __init__(
        self,
        axis=None,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer='zeros',
        gamma_initializer='ones',
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super(InstanceNormalization, self).__init__(**kwargs)
        self.supports_masking = True
        self.axis = axis
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        """

        Parameters
        ----------
        input_shape : tuple

        """
        ndim = len(input_shape)
        if self.axis == 0:
            raise ValueError('Axis cannot be zero')

        if (self.axis is not None) and (ndim == 2):
            raise ValueError('Cannot specify axis for rank 1 tensor')

        self.input_spec = InputSpec(ndim=ndim)

        if self.axis is None:
            shape = (1,)
        else:
            shape = (input_shape[self.axis],)

        if self.scale:
            self.gamma = self.add_weight(
                shape=shape,
                name='gamma',
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
            )
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(
                shape=shape,
                name='beta',
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
            )
        else:
            self.beta = None
        self.built = True

    def call(self, inputs, training=None):
        """

        Parameters
        ----------
        inputs
        training

        Returns
        -------

        """
        input_shape = backend.int_shape(inputs)
        reduction_axes = list(range(0, len(input_shape)))

        if self.axis is not None:
            del reduction_axes[self.axis]

        del reduction_axes[0]

        mean = backend.mean(inputs, reduction_axes, keepdims=True)
        stddev = backend.std(inputs, reduction_axes, keepdims=True) + self.epsilon
        normed = (inputs - mean) / stddev

        broadcast_shape = [1] * len(input_shape)
        if self.axis is not None:
            broadcast_shape[self.axis] = input_shape[self.axis]

        if self.scale:
            broadcast_gamma = backend.reshape(self.gamma, broadcast_shape)
            normed = normed * broadcast_gamma
        if self.center:
            broadcast_beta = backend.reshape(self.beta, broadcast_shape)
            normed = normed + broadcast_beta
        return normed

    def get_config(self):
        """

        Returns
        -------
        config dict : dict

        """
        config = {
            'axis': self.axis,
            'epsilon': self.epsilon,
            'center': self.center,
            'scale': self.scale,
            'beta_initializer': initializers.serialize(self.beta_initializer),
            'gamma_initializer': initializers.serialize(self.gamma_initializer),
            'beta_regularizer': regularizers.serialize(self.beta_regularizer),
            'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
            'beta_constraint': constraints.serialize(self.beta_constraint),
            'gamma_constraint': constraints.serialize(self.gamma_constraint),
        }
        base_config = super(InstanceNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from functools import reduce
from operator import mul
from typing import Tuple, Union, Sequence

import numexpr
import numpy
from numba import jit, cuda
from numba.cuda import CudaSupportError

from aydin.features.groups.base import FeatureGroupBase


from aydin.util.array.nd import nd_range_radii
from aydin.util.fast_shift.fast_shift import fast_shift
from aydin.util.fast_uniform_filter.numba_cpu_uf import numba_cpu_uniform_filter
from aydin.util.fast_uniform_filter.parallel_uf import parallel_uniform_filter
from aydin.util.log.log import lprint


# Removes duplicates without changing list's order:
def _remove_duplicates(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


class UniformFeatures(FeatureGroupBase):
    """
    Uniform Feature Group class
    """

    def __init__(
        self,
        kernel_widths=None,
        kernel_scales=None,
        kernel_shapes=None,
        min_level=0,
        max_level=13,
        include_scale_one=False,
        include_fine_features=True,
        include_corner_features=False,
        include_line_features=False,
        decimate_large_scale_features=True,
        extend_large_scale_features=False,
        scale_one_width=3,
        dtype=numpy.float32,
    ):
        super().__init__()

        # Setting up default features:
        #  0,  1,  2,  3,  4,  5,  6,   7,   8,    9,   10,   11
        if kernel_widths is None:
            kernel_widths = []
            if include_scale_one:
                kernel_widths += [scale_one_width]
            if include_fine_features:
                kernel_widths += [3, 3, 3]
            if include_line_features:
                kernel_widths += [3, 3, 3]
            kernel_widths += [3] * 10
        if kernel_scales is None:
            kernel_scales = []
            if include_scale_one:
                kernel_scales += [1]
            if include_fine_features:
                kernel_scales += [3, 5, 7]
            if include_line_features:
                kernel_scales += [3, 5, 7]
            kernel_scales += [2 ** i - 1 for i in range(2, 12)]
        if kernel_shapes is None:
            kernel_shapes = []
            if include_scale_one:
                kernel_shapes += ['li']
            if include_fine_features:
                special_feature_shape = '#linc' if include_corner_features else '#l1nc'
                kernel_shapes += [special_feature_shape] * 3
            if include_line_features:
                kernel_shapes += ['|l1'] * 3

            prefix = '*' if extend_large_scale_features else ''

            kernel_shapes += [prefix + 'l2'] * 3

            if decimate_large_scale_features:
                kernel_shapes += [prefix + 'l1nc'] * 4 + [prefix + 'l1oc'] * 3
            else:
                kernel_shapes += [prefix + 'l1'] * 4 + [prefix + 'l1'] * 3

        self.kernel_widths = kernel_widths
        self.kernel_scales = kernel_scales
        self.kernel_shapes = kernel_shapes

        self.min_level = min_level
        self.max_level = max_level

        self.dtype = dtype

        self._feature_descriptions_list = None
        self._size_to_full_feature = None

        self.image = None
        self.gpu_image = None
        self.cuda_stream = None
        self.original_dtype = None
        self.excluded_voxels: Sequence[Tuple[int, ...]] = []
        self.args = None
        self.kwargs = None

    def _ensure_feature_description_available(self, ndim: int):
        if (
            self._feature_descriptions_list is None
            or len(self._feature_descriptions_list[0][0]) != ndim
        ):
            self._feature_descriptions_list = self._get_feature_descriptions_list(ndim)

    def _get_feature_descriptions_list(self, ndim: int):
        """
        Get feature descriptions

        Parameters
        ----------
        ndim : int

        Returns
        -------
        feature_description_list

        """
        feature_description_list = []

        level = 0
        for width, scale, shape in zip(
            self.kernel_widths, self.kernel_scales, self.kernel_shapes
        ):
            # Check if we have passed the max number of features already:
            # Important: We might overshoot a bit, but that's ok to make sure we get all features at a given scale...
            if level >= self.max_level:
                break
            elif level < self.min_level:
                level += 1
                continue
            level += 1

            # Computing the radius:
            radius = width // 2

            # We compute the radii along the different dimensions:
            radii = list((max(1, radius),) * ndim)

            # We generate all feature shift vectors:
            features_shifts = list(nd_range_radii(radii))

            # print(f'Feature shifts: {features_shifts}')

            # For each feature shift we append to the feature description list:
            for feature_shift in features_shifts:

                # Excluding the center pixel/feature:
                if scale == 1 and feature_shift == (0,) * ndim:
                    continue

                # if scale == 2 and width == 1:
                #     effective_shift = feature_shift
                #     negative_extent = (0,) * ndim
                #     positive_extent = (2,) * ndim

                # Different 'shapes' of feature  distributions:
                if 'l1' in shape and sum([abs(i) for i in feature_shift]) > radius:
                    continue
                elif (
                    'l2' in shape
                    and sum([i * i for i in feature_shift]) > radius * radius
                ):
                    continue
                elif 'li' in shape:
                    pass

                # keep only center (oc) or remove all that are center (nc)
                if 'oc' in shape and sum([abs(i) for i in feature_shift]) > 0:
                    continue
                elif 'nc' in shape and sum([abs(i) for i in feature_shift]) == 0:
                    continue

                hscale = scale // 2
                if '#' in shape:
                    effective_shift = tuple(0 for _ in feature_shift)
                    negative_extent = tuple(
                        (hscale if s == 0 else hscale * abs(max(0, s)))
                        for d, s in zip(range(ndim), feature_shift)
                    )
                    positive_extent = tuple(
                        (hscale if s == 0 else hscale * abs(min(0, s)))
                        for d, s in zip(range(ndim), feature_shift)
                    )
                elif '|' in shape:
                    effective_shift = tuple(i * hscale for i in feature_shift)
                    negative_extent = tuple(
                        (hscale if s == 0 else abs(max(0, s)))
                        for d, s in zip(range(ndim), feature_shift)
                    )
                    positive_extent = tuple(
                        (hscale if s == 0 else abs(min(0, s)))
                        for d, s in zip(range(ndim), feature_shift)
                    )
                elif '*' in shape:
                    effective_shift = tuple(i * scale for i in feature_shift)
                    negative_extent = tuple(
                        (
                            max(0, hscale * (2 + radius))
                            if s == 0
                            else (abs(s) * scale - 2 if s > 0 else hscale)
                        )
                        for d, s in zip(range(ndim), feature_shift)
                    )
                    positive_extent = tuple(
                        (
                            max(0, hscale * (2 + radius))
                            if s == 0
                            else (abs(s) * scale - 2 if s < 0 else hscale)
                        )
                        for d, s in zip(range(ndim), feature_shift)
                    )
                else:
                    effective_shift = tuple(i * scale for i in feature_shift)
                    negative_extent = (hscale,) * ndim
                    positive_extent = (hscale,) * ndim

                feature_description = (
                    effective_shift,
                    negative_extent,
                    positive_extent,
                    shape,
                )

                # Now we check if the feature overlaps with any excluded voxels:
                # if check_for_excluded_voxels(
                #         effective_shift, negative_extent, positive_extent, excluded_voxels
                # ):
                #     continue

                #  We append the feature description:
                feature_description_list.append(feature_description)

        # Some features might be identical due to the aspect ratio, we eliminate duplicates:
        no_duplicate_feature_description_list = _remove_duplicates(
            feature_description_list
        )
        # We save the last computed feature description list for debug purposes:
        self.debug_feature_description_list = feature_description_list
        # We check and report how many duplicates were eliminated:
        number_of_duplicates = len(feature_description_list) - len(
            no_duplicate_feature_description_list
        )

        lprint(f"Number of duplicate features: {number_of_duplicates}")
        feature_description_list = no_duplicate_feature_description_list
        return feature_description_list

    @property
    def receptive_field_radius(self) -> int:
        radius = 0
        counter = 0
        for width, scale in zip(self.kernel_widths, self.kernel_scales):

            if counter > self.max_level:
                break

            radius = max(radius, width * scale // 2)
            counter += 1

        return radius

    def num_features(self, ndim: int) -> int:
        self._ensure_feature_description_available(ndim)
        return len(self._feature_descriptions_list)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        # Save original image dtype:
        self.original_dtype = image.dtype

        # Scipy does not support float16 yet:
        dtype = (
            numpy.float32
            if self.original_dtype == numpy.float16
            else self.original_dtype
        )
        image = image.astype(dtype=dtype, copy=False)

        self.image = image
        self.excluded_voxels = excluded_voxels
        self.kwargs = kwargs

        # Let's make sure we have the descriptions of the features:
        self._ensure_feature_description_available(image.ndim)

        size_to_feature = {}
        for feature_description in self._feature_descriptions_list:
            # Unpacking the description:
            translation, negative_extent, positive_extent, shape = feature_description

            # Calculating the uniform feature size:
            size = tuple((n + 1 + s for n, s in zip(negative_extent, positive_extent)))

            # Let's check that the feature is not already computed:
            if size not in size_to_feature:
                lprint(f"Pre-computing uniform filter of size: {size}")
                # Compute the feature
                feature = self._compute_uniform_filter(image, size=size)

                # save feature in cache:
                size_to_feature[size] = feature

        self._size_to_full_feature = size_to_feature

    def compute_feature(self, index: int, feature):

        feature_description = self._feature_descriptions_list[index]
        lprint(
            f"Uniform feature: {index}, description: {feature_description}, excluded_voxels={self.excluded_voxels}"
        )

        # Unpacking the description:
        translation, negative_extent, positive_extent, shape = feature_description

        # Calculating the uniform feature size:
        size = tuple((n + 1 + s for n, s in zip(negative_extent, positive_extent)))

        # Fetching the corresponding full feature:
        full_feature = self._size_to_full_feature[size]

        # We use the full uniform filter result and modify it accordingly:
        self._translate_and_exclude_center_value(
            self.image, full_feature, feature, feature_description, self.excluded_voxels
        )

    def _translate_and_exclude_center_value(
        self, image, feature_in, feature_out, feature_description, excluded_voxels
    ):
        """
        It is not recommended to optimise this function as there is some very technical points happening here,
        and the key functions that need optimizing have been externalised anyway (see below).
        """

        # This function exists to facilitate the implementation of optimised versions of it.

        # Unpacking the description:
        translation, negative_extent, positive_extent, shape = feature_description

        # Adjust translation given the extents:
        # There is something a bit tricky here, this term: ' + (p - n) '
        # That's needed to take into account the fact that the center of a
        # uniform filter of even shape does not land on a voxel...
        translation_adjusted = tuple(
            (
                -t + (p - n) // 2
                for t, n, p in zip(translation, negative_extent, positive_extent)
            )
        )

        # We check that the translation is not trivial:
        if any(abs(t) > 0 for t in translation_adjusted):
            self._translate_image(feature_in, feature_out, translation_adjusted)
        else:
            # If the translation does not translated anything then let's just not translate, right?
            feature_out[...] = feature_in

        # We store here the sum of excluded values to be able to substract:
        excluded_values_sum = None
        # And count how many voxels are effectively excluded:
        excluded_count = 0

        for excluded_voxel in excluded_voxels:
            # Is the center voxel within the filter footprint?
            center_value_within_footprint = all(
                -n <= e - t <= p
                for e, t, n, p in zip(
                    excluded_voxel, translation, negative_extent, positive_extent
                )
            )
            if center_value_within_footprint:
                lprint(f"excluded voxel: {excluded_voxel}")

                # We increment the exclusion count:
                excluded_count += 1

                # Just-in-time allocation:
                if excluded_values_sum is None:
                    excluded_values_sum = numpy.zeros_like(image)

                # In this case we remove the excluded value, first we:
                if any(abs(ev) > 0 for ev in excluded_voxel):
                    excluded_values_sum = self._translate_and_add_image(
                        excluded_values_sum, excluded_voxel, image
                    )
                else:
                    _fast_inplace_add(excluded_values_sum, image)

        # We check if we need to touch the feature at all:
        if excluded_count > 0:

            # Calculating the uniform feature size:
            size = tuple((n + 1 + s for n, s in zip(negative_extent, positive_extent)))

            # we compute the volume of the footprint in voxels:
            footprint_volume = prod(size)

            # If we have more than one excluded voxel, we need to be carefull: if the proportion of excluded voxels
            # in a feature becomes too large (> 10%), we need  to exclude that feature entirely, otherwise it will be
            # too much difference between the feature with or without excluded voxels, and thus too confusing for the regressor.
            # Some of this comes from empirical evidence.
            num_of_excluded_voxels = len(excluded_voxels)
            exclude_feature_entirely = num_of_excluded_voxels > 1 and (
                excluded_count / footprint_volume > 0.1
            )

            if exclude_feature_entirely:
                # If all voxels of the footprint are excluded then the whole feature must be zeroed out:
                feature_out[...] = 0
            else:
                # Then we compute the correction factor so that the feature is the average of the remaining voxels:
                _apply_correction(
                    feature_out, excluded_values_sum, footprint_volume, excluded_count
                )

    def _compute_uniform_filter(self, image, size):
        """
        Override this method to provide an accelerated version
        """

        def no_cuda_cpu_mode():
            # No CUDA? we use CPU mode instead:
            # Different methods perform differently based on filter size:
            max_size = max(size) if isinstance(size, tuple) else size
            if max_size > 128:
                # Numba scales well for large filter sizes:
                output = numba_cpu_uniform_filter(image, size=size, mode="nearest")
                lprint(f"Computed filter of size: {size} with Numba")
            else:
                # Scipy parallel is more efficient for small filter sizes:
                output = parallel_uniform_filter(image, size=size, mode="nearest")
                lprint(f"Computed filter of size: {size} with parallel scipy")

            return output

        if image.size < 1024:
            lprint("Image too small, CUDA not needed!")
            output = no_cuda_cpu_mode()
        else:
            try:
                # No point of using CUDA for very small images!
                # Let's try CUDA first:
                # Note: this is not optimal as the image is pushed to GPU every time...
                from aydin.util.fast_uniform_filter.numba_gpu_uf import (
                    numba_gpu_uniform_filter,
                )

                if self.cuda_stream is None:
                    self.cuda_stream = cuda.stream()

                if self.gpu_image is None:
                    image = numpy.ascontiguousarray(image)
                    self.gpu_image = cuda.to_device(image, stream=self.cuda_stream)

                output = numba_gpu_uniform_filter(
                    self.gpu_image,
                    size=size,
                    mode="nearest",
                    cuda_stream=self.cuda_stream,
                )
                lprint(f"Computed filter of size: {size} with CUDA")
            except Exception as e:
                if isinstance(e, CudaSupportError):
                    lprint(
                        "CUDA not supported on this machine, falling back to numba and scipy."
                    )
                else:
                    import sys

                    error_str = (str(sys.exc_info()[0]) + ', and: ' + str(e)).replace(
                        '\n', ', '
                    )
                    lprint(
                        f"Cannot use CUDA for computing uniform filter because of: {error_str}"
                    )
                output = no_cuda_cpu_mode()

        # Ensure correct type:
        dtype = image.dtype if self.dtype is None else self.dtype
        dtype = numpy.float32 if dtype == numpy.float16 else dtype
        output = output.astype(dtype=dtype, copy=False)

        return output

    def _translate_and_add_image(self, sum_image, translation, image):
        """
        Override this method to provide an accelerated version
        """
        fast_shift(image, shift=tuple(translation), output=sum_image, add=True)
        return sum_image

    def _translate_image(self, feature_in, feature_out, translation):
        """
        Override this method to provide an accelerated version
        """
        fast_shift(feature_in, shift=tuple(translation), output=feature_out)

    def finish(self):
        # Here we cleanup any resource alocated for the last feature computation:
        self.image = None
        self.excluded_voxels = None
        self.kwargs = None
        self._feature_descriptions_list = None
        self._size_to_full_feature = None
        self.gpu_image = None
        self.cuda_stream = None


def _apply_correction(
    feature, excluded_values_sum, footprint_volume: int, excluded_count: int
):
    if (
        feature.dtype == numpy.int16
        or feature.dtype == numpy.uint16
        or feature.dtype == numpy.int8
        or feature.dtype == numpy.uint8
    ):
        # feature_float = feature.astype(dtype=numpy.float32, copy=False)
        # _apply_correction_numba(feature_float, excluded_values_sum, footprint_volume, excluded_count)
        # feature[...] = feature_float.astype(dtype=feature.dtype, copy=False)
        alpha = footprint_volume / (footprint_volume - excluded_count)
        beta = -alpha / footprint_volume  # noqa: F841

        numexpr.evaluate(
            "alpha*feature + beta*excluded_values_sum", out=feature, casting='unsafe'
        )
    else:
        _apply_correction_numba(
            feature, excluded_values_sum, footprint_volume, excluded_count
        )


@jit(
    nopython=True,
    # parallel=True,
    error_model='numpy',
    fastmath={'contract', 'afn', 'reassoc'},
)
def _apply_correction_numba(
    feature, excluded_values_sum, footprint_volume: int, excluded_count: int
):
    alpha = footprint_volume / (footprint_volume - excluded_count)
    beta = -alpha / footprint_volume

    feature *= alpha
    feature += beta * excluded_values_sum


@jit(
    nopython=True,
    parallel=True,
    error_model='numpy',
    fastmath={'contract', 'afn', 'reassoc'},
)
def _fast_inplace_add(a, b):
    a += b


def prod(atuple: Tuple[Union[float, int]]):
    # In python 3.8 there is a prod function in math, until then we have:
    return reduce(mul, atuple)



from pprint import pprint

import numpy

from aydin.util.array.outer import outer_product, outer_sum


def test_outer_sum():

    u = numpy.linspace(0, 1, 4)
    v = numpy.linspace(-1, 1, 5) ** 2
    p = outer_sum(u, v)
    print("")
    pprint(p)

    assert p[2, 3] == u[2] + v[3]


def test_outer_product():

    u = numpy.linspace(0, 1, 4)
    v = numpy.linspace(-1, 1, 5) ** 2
    p = outer_product(u, v)
    print("")
    pprint(p)

    assert p[2, 3] == u[2] * v[3]



import numpy
from numpy.fft import fftn, ifftn, fftshift


def compute_relative_blur_kernel(image_clean, image_blurry, size: int = 3):
    """Compute relative blur kernel.

    Parameters
    ----------
    image_clean : numpy.typing.ArrayLike
    image_blurry : numpy.typing.ArrayLike
    size : int

    Returns
    -------
    kernel : numpy.typing.ArrayLike

    """

    image_blurry_fft = fftn(image_blurry)
    image_clean_fft = fftn(image_clean)
    kernel_fft = image_clean_fft / image_blurry_fft
    kernel = ifftn(kernel_fft)

    radius = (size - 1) // 2
    slice_spec = tuple(
        slice(s // 2 - radius, s // 2 + radius + 1) for s in kernel.shape
    )
    kernel = fftshift(kernel)
    kernel = kernel[slice_spec]
    kernel = numpy.absolute(kernel)
    kernel = kernel / kernel.sum()

    return kernel



import itertools
import math
from functools import partial

import numpy
import numpy as np
from numpy import zeros_like
from scipy import ndimage as ndi
from scipy.ndimage import convolve
from skimage.metrics import mean_squared_error

from aydin.analysis.blind_spot_analysis import auto_detect_blindspots
from aydin.util.log.log import lsection, lprint


def calibrate_denoiser_classic(
    image,
    denoise_function,
    denoise_parameters,
    *,
    stride=4,
    loss_function=mean_squared_error,
    # _structural_loss, mean_squared_error, mean_absolute_error #
    display_images=False,
    **other_fixed_parameters,
):
    """
    Calibrates denoiser using self-supervised loss from Batson & Royer*
    Derived from code here:
    https://scikit-image.org/docs/dev/auto_examples/filters/plot_j_invariant_tutorial.html
    Reference: "Noise2Self: Blind Denoising by Self-Supervision, International
    Conference on Machine Learning, p. 524-533 (2019)"

    This 'classic_denoisers' version uses a 'brute-force' optimizer. Good when the
    denoiser is fast enough and the parameter space to explore small enough.

    Parameters
    ----------
    image: ArrayLike
        Image to calibate denoiser with.
    denoise_function: Callable
        Denosing function to calibrate. Should take an image as first parameter,
        all other parameters should have defaults
    denoise_parameters:
        Dictionary with keys corresponding to parameters of the denoising function.
        Values are either: (i) a list of possible values (categorical parameter),
        or (ii) a tuple of floats defining the bounds of that numerical parameter.
    stride: int
        Stride to compute self-supervised loss.
    loss_function: Callable
        Loss/Error function: takes two arrays and returns a distance-like function.
        Can be:  structural_error, mean_squared_error, _mean_absolute_error
    display_images: bool
        If True the denoised images for each parameter tested are displayed.
        this _will_ be slow.
    other_fixed_parameters: dict
        Other fixed parameters to pass to the denoiser function.


    Returns
    -------
    Dictionary with optimal parameters

    """

    lprint(f"Calibrating denoiser on image of shape: {image.shape}")
    lprint(f"Stride for Noise2Self loss: {stride}")
    lprint(f"Fixed parameters: {other_fixed_parameters}")

    # Pass fixed parameters:
    denoise_function = partial(denoise_function, **other_fixed_parameters)

    parameters_tested, losses = _calibrate_denoiser_search(
        image,
        denoise_function,
        denoise_parameters=denoise_parameters,
        stride=stride,
        loss_function=loss_function,
        display_images=display_images,
    )

    idx = np.argmin(losses)
    best_parameters = parameters_tested[idx]

    lprint(f"Best parameters are: {best_parameters}")

    return best_parameters | other_fixed_parameters


def _j_invariant_loss(
    image,
    denoise_function,
    mask,
    loss_function=mean_squared_error,  # _structural_loss, #
    denoiser_kwargs=None,
):
    image = image.astype(dtype=numpy.float32, copy=False)

    denoised = _invariant_denoise(
        image,
        denoise_function=denoise_function,
        mask=mask,
        denoiser_kwargs=denoiser_kwargs,
    )

    loss = loss_function(image[mask], denoised[mask])

    return loss


def _invariant_denoise(image, denoise_function, mask, *, denoiser_kwargs=None):

    image = image.astype(dtype=numpy.float32, copy=False)

    if denoiser_kwargs is None:
        denoiser_kwargs = {}

    interp = _interpolate_image(image, mask=mask)
    output = np.zeros_like(image)

    input_image = image.copy()
    input_image[mask] = interp[mask]
    output[mask] = denoise_function(input_image, **denoiser_kwargs)[mask]

    return output


def _interpolate_image(image, mask, num_iterations: int = 1):

    # if isinstance(mask, numpy.ndarray):
    #     interpolation = uniform_filter_auto(image, size=3)
    #     if num_iterations > 1:
    #         for _ in range(num_iterations-1):
    #             interpolation[~mask] = image
    #             interpolation = uniform_filter_auto(interpolation)
    # else:
    #     interpolation = uniform_filter_auto(image, size=3)

    conv_filter = ndi.generate_binary_structure(image.ndim, 1).astype(image.dtype)
    conv_filter.ravel()[conv_filter.size // 2] = 0
    conv_filter /= conv_filter.sum()

    interpolation = ndi.convolve(image, conv_filter, mode='mirror')

    return interpolation


def _generate_mask(image, stride: int = 4, max_range: int = 4):

    # Generate slice for mask:
    spatialdims = image.ndim
    n_masks = stride ** spatialdims
    mask = _generate_grid_slice(
        image.shape[:spatialdims], offset=n_masks // 2, stride=stride
    )

    # Do we have to extend these spots?
    blind_spots, noise_auto = auto_detect_blindspots(image, max_range=max_range)
    extended_blind_spot = len(blind_spots) > 1

    if extended_blind_spot:
        # If yes, we need to chnage the way we store the mask:
        mask_full = zeros_like(image, dtype=numpy.bool_)
        mask_full[mask] = True
        mask = mask_full

        spot_kernel = zeros_like(noise_auto, dtype=numpy.bool_)
        for blind_spot in blind_spots:
            blind_spot = tuple(
                slice(s // 2 + x, s // 2 + x + 1)
                for s, x in zip(spot_kernel.shape, blind_spot)
            )
            spot_kernel[blind_spot] = True

        # We extend the spots:
        mask = convolve(mask, spot_kernel)

    return mask


def _generate_grid_slice(shape, *, offset, stride=3):
    phases = np.unravel_index(offset, (stride,) * len(shape))
    mask = tuple(slice(p, None, stride) for p in phases)
    return mask


def _mid_point(numerical_parameters_bounds):
    mid_point = tuple(0.5 * (u + v) for u, v in numerical_parameters_bounds)
    mid_point = numpy.array(mid_point)
    return mid_point


def _product_from_dict(dictionary):
    """Utility function to convert parameter ranges to parameter combinations.

    Converts a dict of lists into a list of dicts whose values consist of the
    cartesian product of the values in the original dict.

    Parameters
    ----------
    dictionary : dict of lists
        Dictionary of lists to be multiplied.

    Yields
    ------
    selections : dicts of values
        Dicts containing individual combinations of the values in the input
        dict.
    """
    keys = dictionary.keys()
    for element in itertools.product(*dictionary.values()):
        yield dict(zip(keys, element))


def _calibrate_denoiser_search(
    image,
    denoise_function,
    denoise_parameters,
    *,
    stride=4,
    loss_function=mean_squared_error,  # _structural_loss, #
    display_images: bool = False,
):
    """Return a parameter search history with losses for a denoise function.

    Parameters
    ----------
    image : ndarray
        Input data to be denoised (converted using `img_as_float`).
    denoise_function : function
        Denoising function to be calibrated.
    denoise_parameters : dict of list
        Ranges of parameters for `denoise_function` to be calibrated over.
    stride : int, optional
        Stride used in masking procedure that converts `denoise_function`
        to J-invariance.


    Returns
    -------
    parameters_tested : list of dict
        List of parameters tested for `denoise_function`, as a dictionary of
        kwargs.
    losses : list of int
        Self-supervised loss for each set of parameters in `parameters_tested`.
    """
    # image = img_as_float(image)
    parameters_tested = list(_product_from_dict(denoise_parameters))
    losses = []
    denoised_images = []

    with lsection(
        f"Searching for best denoising parameters among: {denoise_parameters}"
    ):

        # Generate mask:
        mask = _generate_mask(image, stride)

        for denoiser_kwargs in parameters_tested:
            with lsection(f"computing J-inv loss for: {denoiser_kwargs}"):

                # denoised = _invariant_denoise(
                #     image, denoise_function, mask, denoiser_kwargs=denoiser_kwargs
                # )
                # loss = loss_function(image[mask], denoised[mask])

                # We compute the J-inv loss:
                loss = _j_invariant_loss(
                    image,
                    denoise_function,
                    mask=mask,
                    loss_function=loss_function,
                    denoiser_kwargs=denoiser_kwargs,
                )

                if math.isnan(loss) or math.isinf(loss):
                    loss = math.inf
                lprint(f"J-inv loss is: {loss}")
                losses.append(loss)
                if display_images and not (math.isnan(loss) or math.isinf(loss)):
                    denoised = denoise_function(image, **denoiser_kwargs)
                    denoised_images.append(denoised)

    if display_images:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(numpy.stack(denoised_images), name='denoised')

    return parameters_tested, losses
